# pylint: disable=missing-module-docstring
import numpy as np
import pandas as pd
import cvxpy as cp
import matplotlib.pyplot as plt
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators


class MeanVarianceOptimisation:
    # pylint: disable=too-many-instance-attributes
    """
    This class implements some classic mean-variance optimisation techniques for calculating the efficient frontier solutions.
    With the help of quadratic optimisers, users can generate optimal portfolios for different objective functions. Currently
    solutions to the following portfolios can be generated:

        1. Inverse Variance
        2. Maximum Sharpe
        3. Minimum Volatility
        4. Efficient Risk
        5. Maximum Return - Minimum Volatility
        6. Efficient Return
        7. Maximum Diversification
        8. Maximum Decorrelation
        9. Custom Objective Function
    """

    def __init__(self, calculate_expected_returns='mean', risk_free_rate=0.03):
        """
        Constructor.

        :param calculate_expected_returns: (str) The method to use for calculation of expected returns.
                                                 Currently supports: ``mean``, ``exponential``.
        """

        self.weights = list()
        self.asset_names = None
        self.num_assets = None
        self.portfolio_risk = None
        self.portfolio_return = None
        self.portfolio_sharpe_ratio = None
        self.calculate_expected_returns = calculate_expected_returns
        self.returns_estimator = ReturnsEstimators()
        self.risk_estimators = RiskEstimators()
        self.weight_bounds = (0, 1)
        self.risk_free_rate = risk_free_rate

    def allocate(self, asset_names=None, asset_prices=None, expected_asset_returns=None, covariance_matrix=None,
                 solution='inverse_variance', target_return=0.2, target_risk=0.01, risk_aversion=10, weight_bounds=None):
        # pylint: disable=invalid-name, too-many-branches
        """
        Calculate the portfolio asset allocations using the method specified.

        :param asset_names: (list) A list of strings containing the asset names.
        :param asset_prices: (pd.DataFrame) A dataframe of historical asset prices (daily close).
        :param expected_asset_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param covariance_matrix: (pd.DataFrame/numpy matrix) User supplied covariance matrix of asset returns (sigma).
        :param solution: (str) The type of solution/algorithm to use to calculate the weights.
                               Supported solution strings - ``inverse_variance``, ``min_volatility``, ``max_sharpe``,
                               ``efficient_risk``, ``max_return_min_volatility``, ``max_diversification``, ``efficient_return``
                               and ``max_decorrelation``.
        :param target_return: (float) Target return of the portfolio.
        :param target_risk: (float) Target risk of the portfolio.
        :param risk_aversion: (float) Quantifies the risk averse nature of the investor - a higher value means
                                      more risk averse and vice-versa.
        :param weight_bounds: (dict/tuple) Can be either a single tuple of upper and lower bounds
                                           for all portfolio weights or a list of strings with each string representing
                                           an inequality on the weights. For e.g. to bound the weight of the 3rd asset
                                           pass the following weight bounds: ['weights[2] <= 0.3', 'weights[2] >= 0.1'].
        """

        self._error_checks(asset_names, asset_prices, expected_asset_returns, covariance_matrix, solution)

        # Weight bounds
        if weight_bounds is not None:
            self.weight_bounds = weight_bounds

        # Calculate the expected asset returns and covariance matrix if not given by the user
        expected_asset_returns, covariance = self._calculate_estimators(asset_prices,
                                                                        expected_asset_returns,
                                                                        covariance_matrix)

        if solution == 'inverse_variance':
            self._inverse_variance(covariance=covariance, expected_returns=expected_asset_returns)
        elif solution == 'min_volatility':
            self._min_volatility(covariance=covariance,
                                 expected_returns=expected_asset_returns)
        elif solution == 'max_return_min_volatility':
            self._max_return_min_volatility(covariance=covariance,
                                            expected_returns=expected_asset_returns,
                                            risk_aversion=risk_aversion)
        elif solution == 'max_sharpe':
            self._max_sharpe(covariance=covariance,
                             expected_returns=expected_asset_returns)
        elif solution == 'efficient_risk':
            self._min_volatility_for_target_return(covariance=covariance,
                                                   expected_returns=expected_asset_returns,
                                                   target_return=target_return)
        elif solution == 'efficient_return':
            self._max_return_for_target_risk(covariance=covariance,
                                             expected_returns=expected_asset_returns,
                                             target_risk=target_risk)
        elif solution == 'max_diversification':
            self._max_diversification(covariance=covariance,
                                      expected_returns=expected_asset_returns)
        else:
            self._max_decorrelation(covariance=covariance,
                                    expected_returns=expected_asset_returns)

        # Calculate the portfolio sharpe ratio
        self.portfolio_sharpe_ratio = ((self.portfolio_return - self.risk_free_rate) / (self.portfolio_risk ** 0.5))

        # Do some post-processing of the weights
        self._post_process_weights()

    def allocate_custom_objective(self, non_cvxpy_variables, cvxpy_variables, objective_function, constraints=None):
        # pylint: disable=eval-used, exec-used
        """
        Create a portfolio using custom objective and constraints.

        :param non_cvxpy_variables: (dict) A dictionary of variables to be used for providing the required input matrices and
                                           other inputs required by the user. The key of dictionary will be the variable name
                                           while the value can be anything ranging from a numpy matrix, list, dataframe or number.
        :param cvxpy_variables: (list) This is a list of cvxpy specific variables that will be initialised in the format required
                                       by cvxpy. For e.g. ["risk = cp.quad_form(weights, covariance)"] where you are initialising
                                       a variable named "risk" using cvxpy. Note that cvxpy is being imported as "cp", so be sure
                                       to refer to cvxpy as cp.
        :param custom_objective: (str)  A custom objective function. You need to write it in the form
                                        expected by cvxpy. The objective will be a single string, e.g. 'cp.Maximise(
                                        expected_asset_returns)'.
        :param constraints: (list) a list of strings containing the optimisation constraints. For e.g. ['weights >= 0', 'weights <= 1']
        """

        # Initialise the non-cvxpy variables
        locals_ptr = locals()
        for variable_name, variable_value in non_cvxpy_variables.items():
            exec(variable_name + " = None")
            locals_ptr[variable_name] = variable_value

        self.num_assets = locals_ptr['num_assets']
        self.asset_names = list(range(self.num_assets))
        if 'asset_names' in locals_ptr:
            self.asset_names = locals_ptr['asset_names']

        # Optimisation weights
        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)

        # Initialise cvxpy specific variables
        for variable in cvxpy_variables:
            exec(variable)

        # Optimisation objective and constraints
        allocation_objective = eval(objective_function)
        allocation_constraints = []
        for constraint in constraints:
            allocation_constraints.append(eval(constraint))

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve(warm_start=True)
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')
        self.weights = weights.value

        # Calculate portfolio metrics
        if 'risk' in locals_ptr:
            self.portfolio_risk = locals_ptr['risk'].value
        if 'portfolio_return' in locals_ptr:
            self.portfolio_return = locals_ptr['portfolio_return'].value

        # Do some post-processing of the weights
        self._post_process_weights()

    def get_portfolio_metrics(self):
        """
        Prints the portfolio metrics - return, risk and Sharpe Ratio.
        """

        print("Portfolio Return = %s" % self.portfolio_return)
        print("Portfolio Risk = %s" % self.portfolio_risk)
        print("Portfolio Sharpe Ratio = %s" % self.portfolio_risk)

    def plot_efficient_frontier(self, covariance, expected_asset_returns, min_return=0, max_return=0.4,
                                risk_free_rate=0.05):
        # pylint: disable=broad-except
        """
        Plot the Markowitz efficient frontier.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_asset_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param min_return: (float) Minimum target return.
        :param max_return: (float) Maximum target return.
        :param risk_free_rate: (float) The rate of return for a risk-free asset.
        """

        expected_returns = np.array(expected_asset_returns).reshape((len(expected_asset_returns), 1))
        volatilities = []
        returns = []
        sharpe_ratios = []
        for portfolio_return in np.linspace(min_return, max_return, 100):
            try:
                self.allocate(covariance_matrix=covariance,
                              expected_asset_returns=expected_returns,
                              solution='efficient_risk',
                              target_return=portfolio_return)
                volatilities.append(self.portfolio_risk ** 0.5)
                returns.append(portfolio_return)
                sharpe_ratios.append((portfolio_return - risk_free_rate) / (self.portfolio_risk ** 0.5 + 1e-16))
            except Exception:
                continue
        max_sharpe_ratio_index = sharpe_ratios.index(max(sharpe_ratios))
        min_volatility_index = volatilities.index(min(volatilities))
        figure = plt.scatter(volatilities, returns, c=sharpe_ratios, cmap='viridis')
        plt.colorbar(label='Sharpe Ratio')
        plt.scatter(volatilities[max_sharpe_ratio_index],
                    returns[max_sharpe_ratio_index],
                    marker='*',
                    color='g',
                    s=400,
                    label='Maximum Sharpe Ratio')
        plt.scatter(volatilities[min_volatility_index],
                    returns[min_volatility_index],
                    marker='*',
                    color='r',
                    s=400,
                    label='Minimum Volatility')
        plt.xlabel('Volatility')
        plt.ylabel('Return')
        plt.legend(loc='upper left')
        return figure

    def _error_checks(self, asset_names, asset_prices, expected_asset_returns, covariance_matrix, solution=None):
        """
        Some initial error checks on the inputs.

        :param asset_names: (list) A list of strings containing the asset names.
        :param asset_prices: (pd.DataFrame) A dataframe of historical asset prices (daily close).
        :param expected_asset_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param covariance_matrix: (pd.DataFrame/numpy matrix) User supplied covariance matrix of asset returns (sigma).
        :param solution: (str) The type of solution/algorithm to use to calculate the weights.
                               Currently supported solution strings - inverse_variance, min_volatility, max_sharpe,
                               efficient_risk, max_return_min_volatility, max_diversification, efficient_return
                               and max_decorrelation.
        """

        if asset_prices is None and (expected_asset_returns is None or covariance_matrix is None):
            raise ValueError("You need to supply either raw prices or expected returns "
                             "and a covariance matrix of asset returns")

        if asset_prices is not None:
            if not isinstance(asset_prices, pd.DataFrame):
                raise ValueError("Asset prices matrix must be a dataframe")
            if not isinstance(asset_prices.index, pd.DatetimeIndex):
                raise ValueError("Asset prices dataframe must be indexed by date.")

        if solution is not None and solution not in {"inverse_variance", "min_volatility", "max_sharpe", "efficient_risk",
                                                     "max_return_min_volatility", "max_diversification", "efficient_return",
                                                     "max_decorrelation"}:
            raise ValueError("Unknown solution string specified. Supported solutions - "
                             "inverse_variance, min_volatility, max_sharpe, efficient_risk"
                             "max_return_min_volatility, max_diversification, efficient_return and max_decorrelation")

        if asset_names is None:
            if asset_prices is not None:
                asset_names = asset_prices.columns
            elif covariance_matrix is not None and isinstance(covariance_matrix, pd.DataFrame):
                asset_names = covariance_matrix.columns
            else:
                raise ValueError("Please provide a list of asset names")
        self.asset_names = asset_names
        self.num_assets = len(asset_names)

    def _calculate_estimators(self, asset_prices, expected_asset_returns, covariance_matrix):
        """
        Calculate the expected returns and covariance matrix of assets in the portfolio.

        :param asset_prices: (pd.DataFrame) A dataframe of historical asset prices (daily close).
        :param expected_asset_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param covariance_matrix: (pd.DataFrame/numpy matrix) User supplied covariance matrix of asset returns (sigma).
        :return: (np.array, pd.DataFrame) Expected asset returns and covariance matrix.
        """

        # Calculate the expected returns if the user does not supply any returns
        if expected_asset_returns is None:
            if self.calculate_expected_returns == "mean":
                expected_asset_returns = self.returns_estimator.calculate_mean_historical_returns(
                    asset_prices=asset_prices)
            elif self.calculate_expected_returns == "exponential":
                expected_asset_returns = self.returns_estimator.calculate_exponential_historical_returns(
                    asset_prices=asset_prices)
            else:
                raise ValueError("Unknown returns specified. Supported returns - mean, exponential")
        expected_asset_returns = np.array(expected_asset_returns).reshape((len(expected_asset_returns), 1))

        # Calculate covariance of returns or use the user specified covariance matrix
        if covariance_matrix is None:
            returns = self.returns_estimator.calculate_returns(asset_prices=asset_prices)
            covariance_matrix = returns.cov()
        cov = pd.DataFrame(covariance_matrix, index=self.asset_names, columns=self.asset_names)

        return expected_asset_returns, cov

    def _post_process_weights(self):
        """
        Check weights for very small numbers and numbers close to 1. A final post-processing of weights produced by the
        optimisation procedures.
        """

        # Round weights which are very very small negative numbers (e.g. -4.7e-16) to 0
        self.weights[self.weights < 0] = 0

        # If any of the weights is very close to one, we convert it to 1 and set the other asset weights to 0.
        if True in set(np.isclose(self.weights, 1)):
            almost_one_index = np.isclose(self.weights, 1)
            self.weights[almost_one_index] = 1
            self.weights[np.logical_not(almost_one_index)] = 0

        self.weights = pd.DataFrame(self.weights)
        self.weights.index = self.asset_names
        self.weights = self.weights.T

    def _inverse_variance(self, covariance, expected_returns):
        """
        Calculate weights using inverse-variance allocation.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        """

        ivp = 1. / np.diag(covariance)
        ivp /= ivp.sum()
        self.weights = ivp
        self.portfolio_risk = np.dot(self.weights, np.dot(covariance.values, self.weights.T))
        self.portfolio_return = np.dot(self.weights, expected_returns)[0]

    def _min_volatility(self, covariance, expected_returns):
        # pylint: disable=eval-used
        """
        Compute minimum volatility portfolio allocation.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        """

        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)
        risk = cp.quad_form(weights, covariance)
        portfolio_return = cp.matmul(weights, expected_returns)

        # Optimisation objective and constraints
        allocation_objective = cp.Minimize(risk)
        allocation_constraints = [
            cp.sum(weights) == 1,
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    weights >= self.weight_bounds[0],
                    weights <= min(self.weight_bounds[1], 1)
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    weights <= 1,
                    weights >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve(warm_start=True)
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = risk.value
        self.portfolio_return = portfolio_return.value[0]

    def _max_return_min_volatility(self, covariance, expected_returns, risk_aversion):
        # pylint: disable=eval-used
        """
        Calculate maximum return-minimum volatility portfolio allocation.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param risk_aversion: (float) Quantifies the risk-averse nature of the investor - a higher value means
                           more risk averse and vice-versa.
        """

        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)
        portfolio_return = cp.matmul(weights, expected_returns)
        risk = cp.quad_form(weights, covariance)

        # Optimisation objective and constraints
        allocation_objective = cp.Minimize(risk_aversion * risk - portfolio_return)
        allocation_constraints = [
            cp.sum(weights) == 1
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    weights >= self.weight_bounds[0],
                    weights <= min(self.weight_bounds[1], 1)
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    weights <= 1,
                    weights >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve(warm_start=True)
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = risk.value
        self.portfolio_return = portfolio_return.value[0]

    def _max_sharpe(self, covariance, expected_returns):
        # pylint: disable=invalid-name, eval-used
        """
        Compute maximum Sharpe portfolio allocation.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        """

        y = cp.Variable(self.num_assets)
        y.value = np.array([1 / self.num_assets] * self.num_assets)
        kappa = cp.Variable(1)
        risk = cp.quad_form(y, covariance)
        weights = y / kappa
        portfolio_return = cp.matmul(weights, expected_returns)

        # Optimisation objective and constraints
        allocation_objective = cp.Minimize(risk)
        allocation_constraints = [
            cp.sum((expected_returns - self.risk_free_rate).T @ y) == 1,
            cp.sum(y) == kappa,
            kappa >= 0
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    y >= kappa * self.weight_bounds[0],
                    y <= kappa * self.weight_bounds[1]
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    y <= kappa,
                    y >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve(warm_start=True)
        if y.value is None or kappa.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = risk.value
        self.portfolio_return = portfolio_return.value[0]

    def _min_volatility_for_target_return(self, covariance, expected_returns, target_return):
        # pylint: disable=eval-used
        """
        Calculate minimum volatility portfolio for a given target return.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param target_return: (float) Target return of the portfolio.
        """

        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)
        risk = cp.quad_form(weights, covariance)
        portfolio_return = cp.matmul(weights, expected_returns)

        # Optimisation objective and constraints
        allocation_objective = cp.Minimize(risk)
        allocation_constraints = [
            cp.sum(weights) == 1,
            portfolio_return >= target_return,
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    weights >= self.weight_bounds[0],
                    weights <= min(self.weight_bounds[1], 1)
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    weights <= 1,
                    weights >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve()
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = risk.value
        self.portfolio_return = target_return

    def _max_return_for_target_risk(self, covariance, expected_returns, target_risk):
        # pylint: disable=eval-used
        """
        Calculate maximum return for a given target volatility/risk.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        :param target_risk: (float) Target risk of the portfolio.
        """

        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)
        portfolio_return = cp.matmul(weights, expected_returns)
        risk = cp.quad_form(weights, covariance)

        # Optimisation objective and constraints
        allocation_objective = cp.Maximize(portfolio_return)
        allocation_constraints = [
            cp.sum(weights) == 1,
            risk <= target_risk
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    weights >= self.weight_bounds[0],
                    weights <= min(self.weight_bounds[1], 1)
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    weights <= 1,
                    weights >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve()
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = target_risk
        self.portfolio_return = portfolio_return.value[0]

    def _max_diversification(self, covariance, expected_returns):
        """
        Calculate the maximum diversified portfolio.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        """

        self._max_decorrelation(covariance, expected_returns)

        # Divide weights by individual asset volatilities
        self.weights /= np.diag(covariance)

        # Standardize weights
        self.weights /= np.sum(self.weights)

        portfolio_return = np.dot(expected_returns.T, self.weights)[0]
        risk = np.dot(self.weights, np.dot(covariance, self.weights.T))

        self.portfolio_risk = risk
        self.portfolio_return = portfolio_return

    def _max_decorrelation(self, covariance, expected_returns):
        # pylint: disable=eval-used
        """
        Calculate the maximum decorrelated portfolio.

        :param covariance: (pd.DataFrame) Covariance dataframe of asset returns.
        :param expected_returns: (list/np.array/pd.dataframe) A list of mean stock returns (mu).
        """

        weights = cp.Variable(self.num_assets)
        weights.value = np.array([1 / self.num_assets] * self.num_assets)
        risk = cp.quad_form(weights, covariance)
        portfolio_return = cp.matmul(weights, expected_returns)
        corr = self.risk_estimators.cov_to_corr(covariance)
        portfolio_correlation = cp.quad_form(weights, corr)

        # Optimisation objective and constraints
        allocation_objective = cp.Minimize(portfolio_correlation)
        allocation_constraints = [
            cp.sum(weights) == 1
        ]
        if isinstance(self.weight_bounds, tuple):
            allocation_constraints.extend(
                [
                    weights >= self.weight_bounds[0],
                    weights <= min(self.weight_bounds[1], 1)
                ]
            )
        else:
            for inequality in self.weight_bounds:
                allocation_constraints.append(eval(inequality))

            # Add the hard-boundaries for weights.
            allocation_constraints.extend(
                [
                    weights <= 1,
                    weights >= 0
                ]
            )

        # Define and solve the problem
        problem = cp.Problem(
            objective=allocation_objective,
            constraints=allocation_constraints
        )
        problem.solve(warm_start=True)
        if weights.value is None:
            raise ValueError('No optimal set of weights found.')

        self.weights = weights.value
        self.portfolio_risk = risk.value
        self.portfolio_return = portfolio_return.value[0]



"""
Varoius codependence measure: mutual info, distance correlations, variation of information
"""

from famlafl.codependence.correlation import (angular_distance, absolute_angular_distance, squared_angular_distance, \
    distance_correlation)
from famlafl.codependence.information import (get_mutual_info, get_optimal_number_of_bins, \
    variation_of_information_score)
from famlafl.codependence.codependence_matrix import (get_dependence_matrix, get_distance_matrix)
from famlafl.codependence.gnpr_distance import (spearmans_rho, gpr_distance, gnpr_distance)



"""
Return in excess of median method.

Described in "The benefits of tree-based models for stock selection", Zhu et al. (2012). Data labeled this way can be
used in regression and classification models to predict stock returns over market.
"""
import numpy as np


def excess_over_median(prices, binary=False, resample_by=None, lag=True):
    """
    Return in excess of median labeling method. Sourced from "The benefits of tree-based models for stock selection"
    Zhu et al. (2012).

    Returns a DataFrame containing returns of stocks over the median of all stocks in the portfolio, or returns a
    DataFrame containing signs of those returns. In the latter case, an observation may be labeled as 0 if it itself is
    the median.

    :param prices: (pd.DataFrame) Close prices of all stocks in the market that are used to establish the median.
                   Returns on each stock are then compared to the median for the given timestamp.
    :param binary: (bool) If False, the numerical value of excess returns over median will be given. If True, then only
                    the sign of the excess return over median will be given (-1 or 1). A label of 0 will be given if
                    the observation itself is the median. According to Zhu et al., categorical labels can alleviate
                    issues with extreme outliers present with numerical labels.
    :param resample_by: (str) If not None, the resampling period for price data prior to calculating returns. 'B' = per
                        business day, 'W' = week, 'M' = month, etc. Will take the last observation for each period.
                        For full details see `here.
                        <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
    :param lag: (bool) If True, returns will be lagged to make them forward-looking.
    :return: (pd.DataFrame) Numerical returns in excess of the market median return, or sign of return depending on
                    whether binary is False or True respectively.
    """
    # Apply resample, if applicable.
    if resample_by is not None:
        prices = prices.resample(resample_by).last()

    # Get return per period.
    if lag:
        returns = prices.pct_change(periods=1).shift(-1)
    else:
        returns = prices.pct_change(periods=1)

    # Calculate median returns for each period as market return.
    market_return = returns.median(axis=1)

    # Calculate excess over market (median) return.
    returns_over_median = returns.sub(market_return, axis=0)

    # If binary is true, returns sign of the return over median instead of the value.
    if binary:
        returns_over_median = returns_over_median.apply(np.sign)

    return returns_over_median



"""
This module contains an implementation of an exponentially weighted moving average based on sample size.
The inspiration and context for this code was from a blog post by writen by Maksim Ivanov:
https://towardsdatascience.com/financial-machine-learning-part-0-bars-745897d4e4ba
"""

# Imports
import numpy as np
from numba import jit, float64, int64


@jit(nopython=True)
def ewma(arr_in: np.ndarray, window: int) -> np.ndarray:  # pragma: no cover
    """
    Exponentially weighted moving average specified by a decay ``window`` to provide better adjustments for
    small windows via:
        y[t] = (x[t] + (1-a)*x[t-1] + (1-a)^2*x[t-2] + ... + (1-a)^n*x[t-n]) /
               (1 + (1-a) + (1-a)^2 + ... + (1-a)^n).

    :param arr_in: (np.ndarray) A single dimensional numpy array
    :param window: (int) The decay window
    :return: (np.ndarray) The EWMA vector, same length / shape as ``arr_in``
    """
    arr_length = arr_in.shape[0]
    ewma_arr = np.empty(arr_length, dtype=np.float64)
    alpha = 2.0 / (window + 1)
    weight = 1.0
    ewma_old = arr_in[0]
    ewma_arr[0] = ewma_old
    for i in range(1, arr_length):
        weight += (1 - alpha)**i
        ewma_old = ewma_old * (1 - alpha) + arr_in[i]
        ewma_arr[i] = ewma_old / weight

    return ewma_arr



"""
Classes derived from Portfolio Optimisation module
"""

from famlafl.portfolio_optimization.cla import CriticalLineAlgorithm
from famlafl.portfolio_optimization.hrp import HierarchicalRiskParity
from famlafl.portfolio_optimization.mean_variance import MeanVarianceOptimisation
from famlafl.portfolio_optimization.herc import HierarchicalEqualRiskContribution
from famlafl.portfolio_optimization.risk_metrics import RiskMetrics
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators
from famlafl.portfolio_optimization.nco import NCO
from famlafl.portfolio_optimization.risk_estimators import RiskEstimators
from famlafl.portfolio_optimization.tic import TIC



"""
Test various functions regarding chapter 4: Return/Time attribution
"""

import os
import unittest

import pandas as pd

from famlafl.filters.filters import cusum_filter
from famlafl.labeling.labeling import get_events, add_vertical_barrier
from famlafl.sample_weights.attribution import get_weights_by_return, get_weights_by_time_decay
from famlafl.util.volatility import get_daily_vol


class TestSampling(unittest.TestCase):
    """
    Test Triple barrier, meta-labeling, dropping rare labels, and daily volatility.
    """

    def setUp(self):
        """
        Set the file path for the sample dollar bars data and get triple barrier events
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/dollar_bar_sample.csv'
        self.data = pd.read_csv(self.path, index_col='date_time')
        self.data.index = pd.to_datetime(self.data.index)

        daily_vol = get_daily_vol(close=self.data['close'], lookback=100)
        cusum_events = cusum_filter(self.data['close'], threshold=0.02)
        vertical_barriers = add_vertical_barrier(t_events=cusum_events, close=self.data['close'],
                                                 num_days=2)

        self.data['side'] = 1
        self.meta_labeled_events = get_events(close=self.data['close'],
                                              t_events=cusum_events,
                                              pt_sl=[4, 4],
                                              target=daily_vol,
                                              min_ret=0.005,
                                              num_threads=3,
                                              vertical_barrier_times=vertical_barriers,
                                              side_prediction=self.data['side'],
                                              verbose=False)

    def test_ret_attribution(self):
        """
        Assert that return attribution length equals triple barrier length, check particular values
        """
        non_nan_meta_labels = self.meta_labeled_events.dropna()
        ret_weights = get_weights_by_return(non_nan_meta_labels, self.data['close'], verbose=False)
        self.assertTrue(ret_weights.shape[0] == non_nan_meta_labels.shape[0])
        self.assertTrue(abs(ret_weights.iloc[0] - 0.781807) <= 1e5)
        self.assertTrue(abs(ret_weights.iloc[3] - 1.627944) <= 1e5)

    def test_time_decay_weights(self):
        """
        Assert that time decay weights length equals triple barrier length, check particular values
        """
        non_nan_meta_labels = self.meta_labeled_events.dropna()
        standard_decay = get_weights_by_time_decay(non_nan_meta_labels, self.data['close'], decay=0.5, verbose=False)
        no_decay = get_weights_by_time_decay(non_nan_meta_labels, self.data['close'], decay=1, verbose=False)
        neg_decay = get_weights_by_time_decay(non_nan_meta_labels, self.data['close'], decay=-0.5, verbose=False)
        converge_decay = get_weights_by_time_decay(non_nan_meta_labels, self.data['close'], decay=0, verbose=False)
        pos_decay = get_weights_by_time_decay(non_nan_meta_labels, self.data['close'], decay=1.5, verbose=False)

        self.assertTrue(standard_decay.shape == no_decay.shape)
        self.assertTrue(standard_decay.shape == neg_decay.shape)
        self.assertTrue(standard_decay.shape == converge_decay.shape)
        self.assertTrue(standard_decay.shape == pos_decay.shape)

        self.assertTrue(standard_decay.iloc[-1] == 1.0)
        self.assertTrue(abs(standard_decay.iloc[0] - 0.582191) <= 1e5)

        self.assertTrue(no_decay.values.all() == 1)  # without decay all weights are 1.0

        # negative decay sets 0 weights to approximately decay * length part of data set
        self.assertTrue(neg_decay[neg_decay == 0].shape[0] == 3)

        # in positive decay, weights decrease with increasing label time
        self.assertTrue(pos_decay.iloc[0] == pos_decay.max())
        self.assertTrue(pos_decay.iloc[-2] >= pos_decay.iloc[-1])

    def test_value_error_raise(self):
        """
        Test seq_bootstrap and ind_matrix functions for raising ValueError on nan values
        """

        with self.assertRaises(AssertionError):
            get_weights_by_return(self.meta_labeled_events, self.data['close'])

        with self.assertRaises(AssertionError):
            get_weights_by_time_decay(self.meta_labeled_events, self.data['close'], decay=0.5)



'''
This module generates  synthetic classification dataset of INFORMED, REDUNDANT, and NOISE explanatory
variables based on the book Machine Learning for Asset Manager (code snippet 6.1)
'''
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification

# pylint: disable=invalid-name
def get_classification_data(n_features=100, n_informative=25, n_redundant=25, n_samples=10000, random_state=0, sigma=.0):
    """
    A function to generate synthetic classification datasets

    :param n_features: (int) Total number of features to be generated (i.e. informative + redundant + noisy).
    :param n_informative: (int) Number of informative features.
    :param n_redundant: (int) Number of redundant features.
    :param n_samples: (int) Number of samples (rows) to be generate.
    :param random_state: (int) Random seed.
    :param sigma: (float) This argument is used to introduce substitution effect to the redundant features in
                     the dataset by adding gaussian noise. The lower the  value of  sigma, the  greater the
                     substitution effect.
    :return: (pd.DataFrame, pd.Series)  X and y as features and labels respectively.
    """
    np.random.seed(random_state)
    X, y = make_classification(n_samples=n_samples, n_features=n_features-n_redundant, n_informative=n_informative,
                               n_redundant=0, shuffle=False, random_state=random_state)
    cols = ['I_'+str(i) for i in range(n_informative)]
    cols += ['N_'+str(i) for i in range(n_features-n_informative-n_redundant)]
    X, y = pd.DataFrame(X, columns=cols), pd.Series(y)
    i = np.random.choice(range(n_informative), size=n_redundant)
    for k, j in enumerate(i):
        X['R_'+str(k)] = X['I_'+str(j)]+np.random.normal(size=X.shape[0])*sigma
    return X, y



"""
Explosiveness tests: Chow-Type Dickey-Fuller Test
"""

import pandas as pd
from famlafl.structural_breaks.sadf import get_betas
from famlafl.util import mp_pandas_obj


# pylint: disable=invalid-name

def _get_dfc_for_t(series: pd.Series, molecule: list) -> pd.Series:
    """
    Get Chow-Type Dickey-Fuller Test statistics for each index in molecule

    :param series: (pd.Series) Series to test
    :param molecule: (list) Dates to test
    :return: (pd.Series) Statistics for each index from molecule
    """

    dfc_series = pd.Series(index=molecule, dtype='float64')

    for index in molecule:
        series_diff = series.diff().dropna()
        series_lag = series.shift(1).dropna()
        series_lag[:index] = 0  # D_t* indicator: before t* D_t* = 0

        y = series_diff.loc[series_lag.index].values
        x = series_lag.values
        coefs, coef_vars = get_betas(x.reshape(-1, 1), y)
        b_estimate, b_var = coefs[0], coef_vars[0][0]
        dfc_series[index] = b_estimate / (b_var ** 0.5)

    return dfc_series


def get_chow_type_stat(series: pd.Series, min_length: int = 20, num_threads: int = 8, verbose: bool = True) -> pd.Series:
    """
    Multithread implementation of Chow-Type Dickey-Fuller Test, p.251-252

    :param series: (pd.Series) Series to test
    :param min_length: (int) Minimum sample length used to estimate statistics
    :param num_threads: (int): Number of cores to use
    :param verbose: (bool) Flag to report progress on asynch jobs
    :return: (pd.Series) Chow-Type Dickey-Fuller Test statistics
    """
    # Indices to test. We drop min_length first and last values
    molecule = series.index[min_length:series.shape[0] - min_length]
    dfc_series = mp_pandas_obj(func=_get_dfc_for_t,
                               pd_obj=('molecule', molecule),
                               series=series,
                               num_threads=num_threads,
                               verbose=verbose,
                               )
    return dfc_series



# famlafl/tests/test_multiprocess.py
"""
Test various functions regarding multiprocessing utilities.
"""

import unittest
import pandas as pd
import numpy as np
import io
import time
from unittest.mock import patch

from famlafl.util.multiprocess import (
    mp_pandas_obj,
    process_jobs,
    process_jobs_,
    process_jobs_in_batches,
    lin_parts,
    nested_parts,
    report_progress,
)


def simple_function(molecule):
    """Simple function that returns the molecule unchanged."""
    return pd.Series(index=molecule)

def none_function(molecule):
    """Function that returns None."""
    return None

def int_function(molecule):
    """Function that returns a constant integer."""
    return 42

def fail_on_second_call(molecule):
    """Function that succeeds on first call but fails on second."""
    if not hasattr(fail_on_second_call, 'called'):
        fail_on_second_call.called = True
        return pd.Series(index=molecule)
    raise RuntimeError("Second call fails")


def series_function(molecule):
    """Function that returns a Series with molecule as index."""
    return pd.Series(1, index=molecule)


def dataframe_function(molecule):
    """Function that returns a DataFrame with molecule as index."""
    return pd.DataFrame({'value': 1}, index=molecule)


class TestMultiprocess(unittest.TestCase):
    """
    Test multiprocessing functions.
    """

    def setUp(self):
        """
        Set up test data.
        """
        self.index = pd.date_range('2020-01-01', periods=100, freq='D')

    def test_lin_parts(self):
        """
        Test linear partitioning of atoms.
        """
        # Test with num_atoms > num_threads
        parts = lin_parts(100, 10)
        self.assertEqual(len(parts), 11)  # num_threads + 1 parts
        self.assertEqual(parts[0], 0)
        self.assertEqual(parts[-1], 100)

        # Test with num_atoms < num_threads
        parts = lin_parts(5, 10)
        self.assertEqual(len(parts), 6)  # num_atoms + 1 parts
        self.assertEqual(parts[0], 0)
        self.assertEqual(parts[-1], 5)

    def test_nested_parts(self):
        """
        Test nested partitioning of atoms.
        """
        # Test regular nested partitioning
        parts = nested_parts(100, 10)
        self.assertTrue(len(parts) > 1)
        self.assertEqual(parts[0], 0)
        self.assertTrue(all(parts[i] <= parts[i + 1] for i in range(len(parts) - 1)))

        # Test upper triangle
        parts = nested_parts(100, 10, upper_triangle=True)
        self.assertTrue(len(parts) > 1)
        self.assertEqual(parts[0], 0)

    def test_process_jobs_single_thread(self):
        """
        Test processing jobs with a single thread.
        """
        jobs = [
            {'molecule': self.index[:10], 'func': simple_function},
            {'molecule': self.index[10:20], 'func': simple_function},
        ]
        result = process_jobs(jobs, num_threads=1)
        self.assertEqual(len(result), 2)

    def test_process_jobs_multi_thread(self):
        """
        Test processing jobs with multiple threads.
        """
        jobs = [
            {'molecule': self.index[:10], 'func': simple_function},
            {'molecule': self.index[10:20], 'func': simple_function},
        ]
        result = process_jobs(jobs, num_threads=2)
        self.assertEqual(len(result), 2)

    def test_mp_pandas_obj_empty_input(self):
        """
        Test mp_pandas_obj with empty input.
        """
        empty_index = pd.Index([])
        result = mp_pandas_obj(simple_function, ('molecule', empty_index))
        self.assertTrue(isinstance(result, pd.Series))
        self.assertEqual(len(result), 0)

    def test_mp_pandas_obj_series_output(self):
        """
        Test mp_pandas_obj with Series output.
        """
        result = mp_pandas_obj(series_function, ('molecule', self.index), num_threads=1)
        self.assertTrue(isinstance(result, pd.Series))
        self.assertEqual(len(result), len(self.index))
        self.assertTrue(all(result == 1))

    def test_mp_pandas_obj_dataframe_output(self):
        """
        Test mp_pandas_obj with DataFrame output.
        """
        result = mp_pandas_obj(dataframe_function, ('molecule', self.index), num_threads=1)
        self.assertTrue(isinstance(result, pd.DataFrame))
        self.assertEqual(len(result), len(self.index))
        self.assertTrue(all(result['value'] == 1))

    def test_mp_pandas_obj_linear_partition(self):
        """
        Test mp_pandas_obj with linear partitioning.
        """
        result = mp_pandas_obj(simple_function, ('molecule', self.index), lin_mols=True, num_threads=1)
        self.assertEqual(len(result), len(self.index))

    def test_mp_pandas_obj_nested_partition(self):
        """
        Test mp_pandas_obj with nested partitioning.
        """
        result = mp_pandas_obj(simple_function, ('molecule', self.index), lin_mols=False, num_threads=1)
        self.assertEqual(len(result), len(self.index))

    def test_mp_pandas_obj_single_thread(self):
        """
        Test mp_pandas_obj with single thread.
        """
        result = mp_pandas_obj(simple_function, ('molecule', self.index), num_threads=1)
        self.assertEqual(len(result), len(self.index))

    def test_mp_pandas_obj_multi_thread(self):
        """
        Test mp_pandas_obj with multiple threads.
        """
        result = mp_pandas_obj(simple_function, ('molecule', self.index), num_threads=2)
        self.assertEqual(len(result), len(self.index))

    def test_mp_pandas_obj_batches(self):
        """
        Test mp_pandas_obj with multiple batches.
        """
        result = mp_pandas_obj(simple_function, ('molecule', self.index), mp_batches=2, num_threads=1)
        self.assertEqual(len(result), len(self.index))

    def test_process_jobs_empty_jobs(self):
        """
        Test process_jobs with empty jobs list.
        """
        result = process_jobs([])
        self.assertTrue(isinstance(result, pd.Series))
        self.assertEqual(len(result), 0)

    def test_process_jobs_error_handling(self):
        """
        Test process_jobs error handling.
        """
        def error_function(molecule):
            raise ValueError("Test error")

        jobs = [
            {'molecule': self.index[:10], 'func': error_function},
        ]
        result = process_jobs(jobs, num_threads=1)
        # When all jobs fail, process_jobs should return an empty DataFrame with default columns
        self.assertTrue(isinstance(result, pd.DataFrame))
        self.assertEqual(len(result), 0)
        self.assertTrue(all(col in result.columns for col in ['t1', 'pt', 'sl']))

    # 1) Test process_jobs_ (which returns both outputs and failed jobs)
    def test_process_jobs_variant_success_fail(self):
        """
        Test process_jobs_ variant that returns (outputs, failed).
        """
        jobs = [
            {'molecule': self.index[:10], 'func': simple_function},
            {'molecule': self.index[10:20], 'func': none_function},  # returns None => fail
        ]
        outputs, failed = process_jobs_(jobs, task="test_task", num_threads=1)

        # The first job produces a valid (empty) Series, the second job returns None => goes to 'failed'
        self.assertEqual(len(outputs), 1)  # Only one successful output
        self.assertEqual(len(failed), 1)   # One failed job

    def test_process_jobs_variant_multi_thread_fail(self):
        """
        Test process_jobs_ with multiple threads and a failing job.
        """
        def failing_func(molecule):
            raise RuntimeError("Failing job.")

        jobs = [
            {'molecule': self.index[:10], 'func': simple_function},
            {'molecule': self.index[10:20], 'func': failing_func},
        ]
        outputs, failed = process_jobs_(jobs, num_threads=2)
        self.assertEqual(len(outputs), 1)
        self.assertEqual(len(failed), 1)

    # 2) Test process_jobs_in_batches
    def test_process_jobs_in_batches_success(self):
        """
        Test process_jobs_in_batches with multiple batches and success.
        """
        jobs = []
        for i in range(0, 100, 10):
            jobs.append({'molecule': self.index[i:i+10], 'func': simple_function})

        results, failed = process_jobs_in_batches(
            jobs,
            task="test_batch_task",
            num_threads=2,
            batch_size=2,  # process in batches of 2
            verbose=False  # Turn off printing to keep test output clean
        )
        # Results should be concatenated Series or list of Series
        self.assertTrue(isinstance(results, (pd.Series, list)))
        self.assertEqual(len(failed), 0)

    def test_process_jobs_in_batches_partial_fail(self):
        """
        Test process_jobs_in_batches with some jobs failing.
        """
        jobs = []
        # The first batch will succeed, the second batch fails on second job
        for i in range(0, 20, 10):
            jobs.append({'molecule': self.index[i:i+5], 'func': simple_function})
            jobs.append({'molecule': self.index[i+5:i+10], 'func': none_function})  # returns None => fail

        results, failed = process_jobs_in_batches(
            jobs,
            task="test_batch_task",
            num_threads=1,
            batch_size=2,
            verbose=False
        )
        self.assertTrue(len(failed) > 0)
        # Because the successful ones are Series, the final results might be
        # a single concatenated Series or a list of Series:
        if isinstance(results, pd.Series):
            self.assertGreater(len(results), 0)
        else:
            self.assertGreater(len(results), 0)

    # 3) Edge cases for process_jobs
    def test_process_jobs_non_pandas_outputs(self):
        """
        Test process_jobs with a function returning non-pandas (an integer).
        This should preserve the job structure by returning a list of integers.
        """
        jobs = [
            {'molecule': self.index[:10], 'func': int_function},
            {'molecule': self.index[10:20], 'func': int_function},
        ]
        results = process_jobs(jobs, num_threads=1)
        self.assertEqual(len(results), 2)
        self.assertEqual(results[0], 42)
        self.assertEqual(results[1], 42)

    def test_process_jobs_none_output_single_job(self):
        """
        Test process_jobs with a single job returning None.
        """
        jobs = [{'molecule': self.index[:10], 'func': none_function}]
        results = process_jobs(jobs, num_threads=1)
        # Because the only job returned None, outputs is empty => returns default DataFrame
        self.assertTrue(isinstance(results, pd.DataFrame))
        self.assertEqual(len(results), 0)  # no rows
        self.assertTrue(all(col in results.columns for col in ['t1', 'pt', 'sl']))

    def test_process_jobs_partial_outputs(self):
        """
        Test the partial outputs scenario: one job fails in the middle, or returns None,
        so the length of 'outputs' doesn't match the length of jobs.
        """
        jobs = [
            {'molecule': self.index[:10], 'func': fail_on_second_call},  # first call succeeds, second fails
            {'molecule': self.index[10:20], 'func': fail_on_second_call},
        ]
        results = process_jobs(jobs, num_threads=1)
        # Because the second call fails, process_jobs will catch the exception and
        # return an empty DataFrame with default columns
        self.assertTrue(isinstance(results, pd.DataFrame))
        self.assertEqual(len(results), 0)
        self.assertTrue(all(col in results.columns for col in ['t1', 'pt', 'sl']))

    def test_process_jobs_task_none(self):
        """
        Test that if 'task' is None, the default is set from jobs[0]['func'].__name__.
        We'll just ensure no exception is raised and coverage is triggered.
        """
        jobs = [
            {'molecule': self.index[:10], 'func': simple_function},
            {'molecule': self.index[10:20], 'func': simple_function},
        ]
        # Capture stderr to ensure coverage of report_progress
        with patch('sys.stderr', new=io.StringIO()) as fake_stderr:
            results = process_jobs(jobs, task=None, num_threads=2)
            self.assertEqual(len(results), 2)
            # Check that the default task name (simple_function) got printed
            self.assertIn("simple_function", fake_stderr.getvalue())

    # 4) Directly test coverage of report_progress (optional)
    # Usually covered by process_jobs in multi-thread mode, but here's a direct test if needed:
    def test_report_progress_direct(self):
        """
        Test report_progress by calling it directly with a small example.
        """
        from famlafl.util.multiprocess import report_progress
        with patch('sys.stderr', new=io.StringIO()) as fake_stderr:
            t0 = time.time()
            report_progress(job_number=1, num_jobs=2, time0=t0, task="direct_test")
            output = fake_stderr.getvalue()
            self.assertIn("direct_test", output)
            self.assertIn("% done after", output)
            # Also test final call (job_number == num_jobs)
            with patch('sys.stderr', new=io.StringIO()) as fake_stderr2:
                report_progress(job_number=2, num_jobs=2, time0=t0, task="direct_test")
                self.assertIn("done after", fake_stderr2.getvalue())


# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.up import UP
from famlafl.online_portfolio_selection.corn import CORN


class CORNU(UP):
    """
    This class implements the Correlation Driven Nonparametric Learning - Uniform strategy. It is
    reproduced with modification from the following paper:
    `Li, B., Hoi, S.C., & Gopalkrishnan, V. (2011). CORN: Correlation-driven nonparametric
    learning approach for portfolio selection. ACM TIST, 2,
    21:1-21:29. <https://dl.acm.org/doi/abs/10.1145/1961189.1961193>`_

    Correlation Driven Nonparametric Learning Uniform creates W experts that each follow the CORN
    strategy with the same rho value and different window values. After each period, the weights
    are evenly distributed among all the experts to follow the uniform weights allocation method
    for the universal portfolio strategy.
    """

    def __init__(self, window, rho):
        """
        Initializes Correlation Driven Nonparametric Learning Uniform with the given number of
        windows and set rho value.

        :param window: (int) Number of windows to look back for similarity sets. CORN-K generates
                             experts with range of [1, 2, ..., w]. The window ranges typically work well
                             with shorter terms of [1, 7].
        :param rho: (float) Threshold for similarity with range of [-1, 1].
                            Lower rho values will classify more periods as being similar, and higher
                            values will be more strict on identifying a period as similarly correlated.
                            Rho values between [0, 0.2] typically had higher results.
        """
        self.window = window
        self.rho = rho
        self.number_of_experts = int(self.window)
        super().__init__(number_of_experts=self.number_of_experts, weighted='uniform')

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                  'M' for Month. The inputs are based on pandas' resample method.
        """
        # Check that window value is an integer.
        if not isinstance(self.window, int):
            raise ValueError("Window value must be an integer.")

        # Check that window value is at least 1.
        if self.window < 1:
            raise ValueError("Window value must be greater than or equal to 1.")

        # Check that rho is between -1 and 1.
        if self.rho < -1 or self.rho > 1:
            raise ValueError("Rho must be between -1 and 1.")

        super(CORNU, self)._initialize(asset_prices,
                                       weights,
                                       resample_by)

    def _generate_experts(self):
        """
        Generates W experts from window of [1, w] and same rho value.
        """
        # Initialize expert parameters.
        self.expert_params = np.zeros((self.number_of_experts, 2))
        # Assign number of windows to each experts.
        for n_window in range(self.window):
            # Assign expert with parameters (n_window + 1, rho).
            self.expert_params[n_window] = [n_window + 1, self.rho]

        for exp in range(self.number_of_experts):
            param = self.expert_params[exp]
            self.experts.append(CORN(int(param[0]), param[1]))



"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.multi_product.etf_trick import ETFTrick


class TestETFTrick(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Dollar bars
    2. Volume bars
    3. Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        path = project_path + '/test_data'

        self.open_df_path = '{}/open_df.csv'.format(path)
        self.close_df_path = '{}/close_df.csv'.format(path)
        self.alloc_df_path = '{}/alloc_df.csv'.format(path)
        self.costs_df_path = '{}/costs_df.csv'.format(path)
        self.rates_df_path = '{}/rates_df.csv'.format(path)

        self.open_df = pd.read_csv(self.open_df_path, index_col=0, parse_dates=[0])
        self.close_df = pd.read_csv(self.close_df_path, index_col=0, parse_dates=[0])
        self.alloc_df = pd.read_csv(self.alloc_df_path, index_col=0, parse_dates=[0])
        self.costs_df = pd.read_csv(self.costs_df_path, index_col=0, parse_dates=[0])
        self.rates_df = pd.read_csv(self.rates_df_path, index_col=0, parse_dates=[0])

    def test_etf_trick_costs_defined(self):
        """
        Tests in-memory and csv ETF trick implementation, when costs_df is defined
        """

        csv_etf_trick = ETFTrick(self.open_df_path, self.close_df_path, self.alloc_df_path, self.costs_df_path,
                                 self.rates_df_path)
        in_memory_etf_trick = ETFTrick(self.open_df, self.close_df, self.alloc_df, self.costs_df, self.rates_df)
        in_memory_trick_series = in_memory_etf_trick.get_etf_series()
        csv_trick_series_4 = csv_etf_trick.get_etf_series(batch_size=4)
        csv_etf_trick.reset()
        csv_trick_series_100 = csv_etf_trick.get_etf_series(batch_size=100)
        csv_etf_trick.reset()
        csv_trick_series_all = csv_etf_trick.get_etf_series(batch_size=1e6)

        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_4.shape)
        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_100.shape)
        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_all.shape)

        # Value check
        self.assertTrue(abs(in_memory_trick_series.iloc[20] - 0.9933502) < 1e-6)

        # Assert the first value equal to 1
        self.assertTrue(in_memory_trick_series.iloc[0] == 1.0)
        self.assertTrue(csv_trick_series_4.iloc[0] == 1.0)
        self.assertTrue(csv_trick_series_100.iloc[0] == 1.0)
        self.assertTrue(csv_trick_series_all.iloc[0] == 1.0)

        # Assert the last values are equal
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_4.iloc[-1])
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_100.iloc[-1])
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_all.iloc[-1])

        # Assert same values
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_4.values))
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_100.values))
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_all.values))

    def test_etf_trick_rates_not_defined(self):
        """
        Tests in-memory and csv ETF trick implementation, when costs_df is not defined (should be set trivial)
        """

        csv_etf_trick = ETFTrick(self.open_df_path, self.close_df_path, self.alloc_df_path, self.costs_df_path, None)
        in_memory_etf_trick = ETFTrick(self.open_df, self.close_df, self.alloc_df, self.costs_df, None)
        in_memory_trick_series = in_memory_etf_trick.get_etf_series()
        csv_trick_series_4 = csv_etf_trick.get_etf_series(batch_size=4)
        csv_etf_trick.reset()
        csv_trick_series_100 = csv_etf_trick.get_etf_series(batch_size=100)
        csv_etf_trick.reset()
        csv_trick_series_all = csv_etf_trick.get_etf_series(batch_size=1e6)

        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_4.shape)
        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_100.shape)
        self.assertTrue(in_memory_trick_series.shape == csv_trick_series_all.shape)

        # Value check
        self.assertTrue(abs(in_memory_trick_series.iloc[20] - 0.9933372) < 1e-6)

        # Assert the first value equal to 1
        self.assertTrue(in_memory_trick_series.iloc[0] == 1)
        self.assertTrue(csv_trick_series_4.iloc[0] == 1)
        self.assertTrue(csv_trick_series_100.iloc[0] == 1)
        self.assertTrue(csv_trick_series_all.iloc[0] == 1)

        # Assert the last values are equal
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_4.iloc[-1])
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_100.iloc[-1])
        self.assertTrue(in_memory_trick_series.iloc[-1] == csv_trick_series_all.iloc[-1])

        # Assert same values
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_4.values))
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_100.values))
        self.assertTrue(np.all(in_memory_trick_series.values == csv_trick_series_all.values))

    def test_input_exceptions(self):
        """
        Tests input data frames internal checks
        """
        try:
            ETFTrick(dict(), dict(), dict(), self.costs_df_path, None)
        except TypeError:
            pass

        modified_open_df = self.open_df.copy(deep=True)
        # Add extra timestamp, to generate exception on _index_check()
        modified_open_df.loc[pd.Timestamp(2020, 1, 1), :] = 4
        try:
            ETFTrick(modified_open_df, self.close_df, self.alloc_df, self.costs_df, None)
        except ValueError:
            pass

        csv_etf_trick = ETFTrick(self.open_df_path, self.close_df_path, self.alloc_df_path, self.costs_df_path, None)
        try:
            csv_etf_trick.get_etf_series(batch_size=2)
        except ValueError:
            pass



"""
Fractional differentiation is a technique to make a time series stationary but also
retain as much memory as possible.  This is done by differencing by a positive real
number. Fractionally differenced series can be used as a feature in machine learning
process.
Updated to use Numba for performance-critical loops by Vadim Surin
"""

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from statsmodels.tsa.stattools import adfuller
from numba import njit


class FractionalDifferentiation:
    """
    FractionalDifferentiation class encapsulates the functions that can
    be used to compute fractionally differentiated series.

    Optimizations:
    - Vectorized weight computations
    - Use Numba for performance-critical loops
    - Vectorized convolution approach for frac_diff when possible
    """

    @staticmethod
    @njit
    def _get_weights_numba(diff_amt: float, size: int) -> np.ndarray:
        """
        Numba-accelerated version of weight computation.

        :param diff_amt: (float) Differencing amount
        :param size: (int) Length of the series
        :return: (np.ndarray) Weight vector
        """
        weights = np.zeros(size, dtype=np.float64)
        weights[0] = 1.0
        for k in range(1, size):
            weights[k] = -weights[k - 1] * (diff_amt - k + 1) / k
        return weights[::-1]  # reverse the array

    @staticmethod
    def get_weights(diff_amt: float, size: int) -> np.ndarray:
        """
        Compute the vector of weights for fractional differencing.
        Advances in Financial Machine Learning, Chapter 5, section 5.4.2, page 79.

        The helper function generates weights that are used to compute fractionally
        differentiated series. It computes the weights that get used in the computation
        of  fractionally differentiated series. This generates a non-terminating series
        that approaches zero asymptotically. The side effect of this function is that
        it leads to negative drift "caused by an expanding window's added weights"
        (see page 83 AFML)

        When diff_amt is real (non-integer) positive number then it preserves memory.

        The book does not discuss what should be expected if d is a negative real
        number. Conceptually (from set theory) negative d leads to set of negative
        number of elements. And that translates into a set whose elements can be
        selected more than once or as many times as one chooses (multisets with
        unbounded multiplicity) - see http://faculty.uml.edu/jpropp/msri-up12.pdf.

        :param diff_amt : float Differencing amount. When positive real (non-integer), it preserves memory.
        :param size : int Length of the weight vector to generate.
        :return: (np.ndarray) Weight vector for fractional differencing.
        """
        return FractionalDifferentiation._get_weights_numba(diff_amt, size).reshape(-1, 1)


    @staticmethod
    @njit
    def _get_weights_ffd_numba(diff_amt: float, thresh: float, lim: int) -> np.ndarray:
        """
        Numba-accelerated version for fixed-fractional differencing weights.
        :param diff_amt: (float) Differencing amount
        :param thresh: (float) Threshold for minimum weight
        :param lim: (int) Maximum length of the weight vector
        :return: (np.ndarray) Weight vector
        """
        weights = [1.0]
        k = 1
        # The algorithm below executes the iterativetive estimation (section 5.4.2, page 78)
        # The output weights array is of the indicated length (specified by lim)
        ctr = 0
        while True:
            # compute the next weight
            w_ = -weights[-1] * (diff_amt - k + 1) / k
            if np.abs(w_) < thresh:
                break
            weights.append(w_)
            k += 1
            ctr += 1
            if ctr == lim - 1:  # if we have reached the size limit, exit the loop
                break
        # Now, reverse the list, convert into a numpy column vector
        arr = np.array(weights[::-1], dtype=np.float64).reshape(-1, 1)
        return arr
    
    @staticmethod
    def get_weights_ffd(diff_amt, thresh, lim):
        """
        Advances in Financial Machine Learning, Chapter 5, section 5.4.2, page 83.

        The helper function generates weights that are used to compute fractionally
        differentiate dseries. It computes the weights that get used in the computation
        of fractionally differentiated series. The series is of fixed width and same
        weights (generated by this function) can be used when creating fractional
        differentiated series.
        This makes the process more efficient. But the side-effect is that the
        fractionally differentiated series is skewed and has excess kurtosis. In
        other words, it is not Gaussian any more.

        The discussion of positive and negative d is similar to that in get_weights
        (see the function get_weights)

        :param diff_amt: (float) Differencing amount
        :param thresh: (float) Threshold for minimum weight
        :param lim: (int) Maximum length of the weight vector
        :return: (np.ndarray) Weight vector
        """

        return FractionalDifferentiation._get_weights_ffd_numba(diff_amt, thresh, lim)


    @staticmethod
    def frac_diff(series: pd.DataFrame, diff_amt: float, thresh: float = 0.01) -> pd.DataFrame:
        """
        Expanding window fractional differentiation.
        Uses convolution to speed up computation

        Advances in Financial Machine Learning, Chapter 5, section 5.5, page 82.

        References:
        https://www.wiley.com/en-us/Advances+in+Financial+Machine+Learning-p-9781119482086
        https://wwwf.imperial.ac.uk/~ejm/M3S8/Problems/hosking81.pdf
        https://en.wikipedia.org/wiki/Fractional_calculus

        The steps are as follows:
        - Compute weights (this is a one-time exercise)
        - Iteratively apply the weights to the price series and generate output points

        This is the expanding window variant of the fracDiff algorithm
        Note 1: For thresh-1, nothing is skipped
        Note 2: diff_amt can be any positive fractional, not necessarility bounded [0, 1]

        :param series: (pd.DataFrame) A time series that needs to be differenced
        :param diff_amt: (float) Differencing amount
        :param thresh: (float) Threshold or epsilon
        :return: (pd.DataFrame) Differenced series
        """

        # 1. Compute weights for the longest series
        size = series.shape[0]
        weights = FractionalDifferentiation.get_weights(diff_amt, size)[:, 0] # 1D array

        # 2. Determine initial calculations to be skipped based on weight-loss threshold
        cum_w = np.cumsum(abs(weights))
        cum_w /= cum_w[-1]
        skip = np.sum(cum_w < thresh)

        # We'll use convolution: For an expanding window, the value at time i is:
        # (weights[0]*series[i] + weights[1]*series[i-1] + ...).
        # This is essentially a reversed convolution of the series with weights.
        # Since weights are reversed, we can directly apply convolution and then align indices.

        # 3. Apply weights to values
        output_df = {}
        for col in series.columns:
            col_values = series[col].ffill().to_numpy(np.float64)
            # Convolve series with weights (full mode)
            conv_res = np.convolve(col_values, weights, mode='full')[:len(col_values)]
            # conv_res[i] gives the dot product up to i-th element.
            # We only start using values after 'skip', and align with the original index:
            # Original series length: N
            # conv_res length: N + N - 1 = 2N - 1
            # The valid fractional differences start from index (skip-1), basically.
            # But we must ensure we properly slice to get a length == series length.
            # The expanding window result should have length = original series length.
            # conv_res[:N] corresponds to partial sums up to each index.
            # Actually, for an expanding window, the i-th output corresponds to conv_res[i].
            # Start from skip:
            result = np.full_like(col_values, np.nan)
            if skip < len(result):
                result[skip:] = conv_res[skip:skip+len(result)]
            output_df[col] = result

        return pd.DataFrame(output_df, index=series.index, columns=series.columns)



    @staticmethod
    def frac_diff_ffd(series: pd.DataFrame, diff_amt: float, thresh: float = 1e-5) -> pd.DataFrame:
        """
        Fixed-window fractional differencing.
        We can use a rolling window convolution here as well.

        Advances in Financial Machine Learning, Chapter 5, section 5.5, page 83.

        References:

        * https://www.wiley.com/en-us/Advances+in+Financial+Machine+Learning-p-9781119482086
        * https://wwwf.imperial.ac.uk/~ejm/M3S8/Problems/hosking81.pdf
        * https://en.wikipedia.org/wiki/Fractional_calculus

        The steps are as follows:

        - Compute weights (this is a one-time exercise)
        - Iteratively apply the weights to the price series and generate output points

        Constant width window (new solution)
        Note 1: thresh determines the cut-off weight for the window
        Note 2: diff_amt can be any positive fractional, not necessarity bounded [0, 1].

        :param series: (pd.DataFrame) A time series that needs to be differenced
        :param diff_amt: (float) Differencing amount
        :param thresh: (float) Threshold for minimum weight
        :return: (pd.DataFrame) A data frame of differenced series
        """

        # 1) Compute weights for the longest series
        size = series.shape[0]
        weights = FractionalDifferentiation.get_weights_ffd(diff_amt, thresh, size)[:, 0] # 1D array
        width = len(weights)

        # 2) Apply weights to values
        # 2.1) Start by creating a dictionary to hold all the fractionally differenced series
        output_df = {}
        for col in series.columns:
            col_values = series[col].ffill().to_numpy(np.float64)
            # Convolution with a fixed window:
            # The weights are reversed, so we use 'valid' mode to get the correct length
            conv_res = np.convolve(col_values, weights[::-1], mode='valid')
            # 'valid' mode returns length = N - width + 1
            # Align this with the original index:
            result = np.full_like(col_values, np.nan)
            # The first valid fractional difference appears only after width-1 steps:   
            result[width-1:] = conv_res
            output_df[col] = result

        # transform the dictionary into a data frame
        output_df = pd.DataFrame(output_df, index=series.index, columns=series.columns)
        return output_df


def get_weights(diff_amt: float, size: int) -> np.ndarray:
    """ This is a pass-through function """
    return FractionalDifferentiation.get_weights(diff_amt, size)

def frac_diff(series: pd.DataFrame, diff_amt: float, thresh: float = 0.01) -> pd.DataFrame:
    """ This is a pass-through function """
    return FractionalDifferentiation.frac_diff(series, diff_amt, thresh)


def get_weights_ffd(diff_amt: float, thresh: float, lim: int) -> np.ndarray:
    """ This is a pass-through function """
    return FractionalDifferentiation.get_weights_ffd(diff_amt, thresh, lim)


def frac_diff_ffd(series: pd.DataFrame, diff_amt: float, thresh: float = 1e-5) -> pd.DataFrame:
    """
    Advances in Financial Machine Learning, Chapter 5, section 5.5, page 83.

    References:

    * https://www.wiley.com/en-us/Advances+in+Financial+Machine+Learning-p-9781119482086
    * https://wwwf.imperial.ac.uk/~ejm/M3S8/Problems/hosking81.pdf
    * https://en.wikipedia.org/wiki/Fractional_calculus

    The steps are as follows:

    - Compute weights (this is a one-time exercise)
    - Iteratively apply the weights to the price series and generate output points

    Constant width window (new solution)
    Note 1: thresh determines the cut-off weight for the window
    Note 2: diff_amt can be any positive fractional, not necessarity bounded [0, 1].

    :param series: (pd.Series) A time series that needs to be differenced
    :param diff_amt: (float) Differencing amount
    :param thresh: (float) Threshold for minimum weight
    :return: (pd.DataFrame) A data frame of differenced series
    """
    return FractionalDifferentiation.frac_diff_ffd(series, diff_amt, thresh)

def plot_min_ffd(series: pd.DataFrame) -> plt.Axes:
    """
    Advances in Financial Machine Learning, Chapter 5, section 5.6, page 85.

    References:

    * https://www.wiley.com/en-us/Advances+in+Financial+Machine+Learning-p-9781119482086

    This function plots the graph to find the minimum D value that passes the ADF test.

    It allows to determine d - the amount of memory that needs to be removed to achieve
    stationarity. This function covers the case of 0 < d << 1, when the original series is
    "mildly non-stationary."

    The right y-axis on the plot is the ADF statistic computed on the input series downsampled
    to a daily frequency.

    The x-axis displays the d value used to generate the series on which the ADF statistic is computed.

    The left y-axis plots the correlation between the original series (d=0) and the differentiated
    series at various d values.

    Examples on how to interpret the results of this function are available in the corresponding part
    in the book Advances in Financial Machine Learning.

    :param series: (pd.DataFrame) Dataframe that contains a 'close' column with prices to use.
    :return: (plt.AxesSubplot) A plot that can be displayed or used to obtain resulting data.
    """

    results = pd.DataFrame(columns=['adfStat', 'pVal', 'lags', 'nObs', '95% conf', 'corr'])

    # Iterate through d values with 0.1 step
    for d_value in np.linspace(0, 1, 11):
        close_prices = np.log(series[['close']]).resample('1D').last()  # Downcast to daily obs
        close_prices.dropna(inplace=True)

        # Applying fractional differentiation
        differenced_series = frac_diff_ffd(close_prices, diff_amt=d_value, thresh=0.01).dropna()

        # Correlation between the original and the differentiated series
        corr = np.corrcoef(close_prices.loc[differenced_series.index, 'close'],
                           differenced_series['close'])[0, 1]
        # Applying ADF
        differenced_series = adfuller(differenced_series['close'], maxlag=1, regression='c', autolag=None)

        # Results to dataframe
        results.loc[d_value] = list(differenced_series[:4]) + [differenced_series[4]['5%']] + [corr]  # With critical value

    # Plotting
    plot = results[['adfStat', 'corr']].plot(secondary_y='adfStat', figsize=(10, 8))
    mean_conf = results['95% conf'].mean()
    plt.axhline(mean_conf, linewidth=1, color='r', linestyle='dotted')
    plt.axhline(0, linewidth=1, color='g', linestyle='dotted')

    return plot



"""
Volume classification methods (BVC and tick rule)
"""

from scipy.stats import norm
import pandas as pd


def get_bvc_buy_volume(close: pd.Series, volume: pd.Series, window: int = 20) -> pd.Series:
    """
    Calculates the BVC buy volume

    :param close: (pd.Series): Close prices
    :param volume: (pd.Series): Bar volumes
    :param window: (int): Window for std estimation uses in BVC calculation
    :return: (pd.Series) BVC buy volume
    """
    # .apply(norm.cdf) is used to omit Warning for norm.cdf(pd.Series with NaNs)

    return volume * (close.diff() / close.diff().rolling(window=window).std()).apply(norm.cdf)



"""
Return in excess of mean method.

Chapter 5, Machine Learning for Factor Investing, by Coqueret and Guida, (2020).
"""
import numpy as np


def excess_over_mean(prices, binary=False, resample_by=None, lag=True):
    """
    Return in excess of mean labeling method. Sourced from Chapter 5.5.1 of Machine Learning for Factor Investing,
    by Coqueret, G. and Guida, T. (2020).

    Returns a DataFrame containing returns of stocks over the mean of all stocks in the portfolio. Returns a DataFrame
    of signs of the returns if binary is True. In this case, an observation may be labeled as 0 if it itself is the
    mean.

    :param prices: (pd.DataFrame) Close prices of all tickers in the market that are used to establish the mean. NaN
                    values are ok. Returns on each ticker are then compared to the mean for the given timestamp.
    :param binary: (bool) If False, the numerical value of excess returns over mean will be given. If True, then only
                    the sign of the excess return over mean will be given (-1 or 1). A label of 0 will be given if
                    the observation itself equal to the mean.
    :param resample_by: (str) If not None, the resampling period for price data prior to calculating returns. 'B' = per
                        business day, 'W' = week, 'M' = month, etc. Will take the last observation for each period.
                        For full details see `here.
                        <https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects>`_
    :param lag: (bool) If True, returns will be lagged to make them forward-looking.
    :return: (pd.DataFrame) Numerical returns in excess of the market mean return, or sign of return depending on
                whether binary is False or True respectively.
    """
    # Apply resample, if applicable.
    if resample_by is not None:
        prices = prices.resample(resample_by).last()

    # Get return per period.
    if lag:
        returns = prices.pct_change(periods=1).shift(-1)
    else:
        returns = prices.pct_change(periods=1)

    # Calculate median returns for each period as market return.
    market_return = returns.mean(axis=1)

    # Calculate excess over market (median) return.
    returns_over_mean = returns.sub(market_return, axis=0)

    # If binary is true, returns sign of the return over median instead of the value.
    if binary:
        returns_over_mean = returns_over_mean.apply(np.sign)

    return returns_over_mean



"""
Labeling techniques used in financial machine learning.
"""

from famlafl.labeling.labeling import (add_vertical_barrier, apply_pt_sl_on_t1, barrier_touched, drop_labels,
                                        get_bins, get_events)
from famlafl.labeling.trend_scanning import trend_scanning_labels
from famlafl.labeling.tail_sets import TailSetLabels
from famlafl.labeling.fixed_time_horizon import fixed_time_horizon
from famlafl.labeling.matrix_flags import MatrixFlagLabels
from famlafl.labeling.excess_over_median import excess_over_median
from famlafl.labeling.raw_return import raw_return
from famlafl.labeling.return_vs_benchmark import return_over_benchmark
from famlafl.labeling.excess_over_mean import excess_over_mean



"""
Tests the Critical Line Algorithm (CLA).
"""

import unittest
import os
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.cla import CriticalLineAlgorithm
from famlafl.portfolio_optimization.returns_estimators import ReturnsEstimators


class TestCLA(unittest.TestCase):
    # pylint: disable=too-many-public-methods
    """
    Tests different functions of the CLA class.
    """

    def setUp(self):
        """
        Set the file path for the tick data csv.
        """

        project_path = os.path.dirname(__file__)
        data_path = project_path + '/test_data/stock_prices.csv'
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date")

    def test_cla_with_mean_returns(self):
        """
        Test the calculation of CLA turning points using mean returns.
        """

        self.data.iloc[1:10, :] = 40
        self.data.iloc[11:20, :] = 50
        self.data.iloc[21, :] = 100
        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, asset_names=self.data.columns)
        weights = cla.weights.values
        weights[weights <= 1e-15] = 0 # Convert very very small numbers to 0
        for turning_point in weights:
            assert (turning_point >= 0).all()
            assert len(turning_point) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(turning_point), 1)

    def test_cla_with_weight_bounds_as_lists(self):
        """
        Test the calculation of CLA turning points when we pass the weight bounds as a list
        instead of just lower and upper bound value.
        """

        cla = CriticalLineAlgorithm(weight_bounds=([0]*self.data.shape[1], [1]*self.data.shape[1]),
                                    calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, asset_names=self.data.columns)
        weights = cla.weights.values
        weights[weights <= 1e-15] = 0  # Convert very very small numbers to 0
        for turning_point in weights:
            assert (turning_point >= 0).all()
            assert len(turning_point) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(turning_point), 1)

    def test_cla_with_exponential_returns(self):
        """
        Test the calculation of CLA turning points using exponential returns
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="exponential")
        cla.allocate(asset_prices=self.data, asset_names=self.data.columns)
        weights = cla.weights.values
        weights[weights <= 1e-15] = 0 # Convert very very small numbers to 0
        for turning_point in weights:
            assert (turning_point >= 0).all()
            assert len(turning_point) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(turning_point), 1)

    def test_cla_max_sharpe(self):
        """
        Test the calculation of maximum sharpe ratio weights.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='max_sharpe', asset_names=self.data.columns)
        weights = cla.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_cla_min_volatility(self):
        """
        Test the calculation for minimum volatility weights.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        weights = cla.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_cla_efficient_frontier(self):
        """
        Test the calculation of the efficient frontier solution.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='efficient_frontier', asset_names=self.data.columns)
        assert len(cla.efficient_frontier_means) == len(cla.efficient_frontier_sigma) and \
               len(cla.efficient_frontier_sigma) == len(cla.weights.values)
        assert cla.efficient_frontier_sigma[-1] <= cla.efficient_frontier_sigma[0] and \
               cla.efficient_frontier_means[-1] <= cla.efficient_frontier_means[0]  # higher risk = higher return

    def test_lambda_for_no_bounded_weights(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test the computation of lambda when there are no bounded weights.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        data = self.data.cov()
        data = data.values
        x, y = cla._compute_lambda(covar_f_inv=data,
                                   covar_fb=data,
                                   mean_f=cla.expected_returns,
                                   w_b=None,
                                   asset_index=1,
                                   b_i=[[0], [1]])
        assert isinstance(x, float)
        assert isinstance(y, int)

    def test_free_bound_weights(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test the method of freeing bounded weights when free-weights is None.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        x, y = cla._free_bound_weight(free_weights=[1]*(cla.expected_returns.shape[0]+1))
        assert not x
        assert not y

    def test_expected_returns_equals_means(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test for condition when expected returns equal the mean value.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        data = self.data.copy()
        data.iloc[:, :] = 0.02320653
        cla._initialise(asset_prices=data, resample_by='B', expected_asset_returns=None, covariance_matrix=None)
        assert cla.expected_returns[-1, 0] == 1e-5

    def test_lambda_for_zero_matrices(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test the computation of lambda when there are no bounded weights. The method
        should return None, None.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        data = self.data.cov()
        data = data.values
        data[:, :] = 0
        x, y = cla._compute_lambda(covar_f_inv=data,
                                   covar_fb=data,
                                   mean_f=cla.expected_returns,
                                   w_b=None,
                                   asset_index=1,
                                   b_i=[[0], [1]])
        assert not x
        assert not y

    def test_w_for_no_bounded_weights(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test the computation of weights (w) when there are no bounded weights.
        """

        cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
        cla.allocate(asset_prices=self.data, solution='min_volatility', asset_names=self.data.columns)
        data = self.data.cov()
        data = data.values
        x, y = cla._compute_w(covar_f_inv=data,
                              covar_fb=data,
                              mean_f=cla.expected_returns,
                              w_b=None)
        assert isinstance(x, np.ndarray)
        assert isinstance(y, float)

    def test_purge_excess(self):
        # pylint: disable=protected-access,invalid-name
        """
        Test purge number excess for very very small tolerance.
        """

        with self.assertRaises(IndexError):
            cla = CriticalLineAlgorithm(weight_bounds=(0, 1), calculate_expected_returns="mean")
            cla.allocate(asset_prices=self.data, solution='cla_turning_points', asset_names=self.data.columns)
            cla.weights = list(cla.weights.values)
            cla.weights = cla.weights*100
            cla._purge_num_err(tol=1e-18)

    def test_flag_true_for_purge_num_err(self):
        # pylint: disable=protected-access, no-self-use
        """
        Test whether the flag becomes True in the purge num error function.
        """

        cla = CriticalLineAlgorithm()
        cla.weights = [[1]]
        cla.lower_bounds = [100]
        cla.upper_bounds = [1]
        cla.lambdas = [[1]]
        cla.gammas = [[1]]
        cla.free_weights = [[1]]
        cla._purge_num_err(tol=1)
        assert not cla.weights
        assert not cla.lambdas
        assert not cla.gammas

    def test_value_error_for_unknown_solution(self):
        """
        Test ValueError on passing unknown solution string.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm()
            cla.allocate(asset_prices=self.data, solution='unknown_string', asset_names=self.data.columns)

    def test_value_error_for_non_dataframe_input(self):
        """
        Test ValueError on passing non-dataframe input.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm()
            cla.allocate(asset_prices=self.data.values, solution='cla_turning_points', asset_names=self.data.columns)

    def test_value_error_for_non_date_index(self):
        """
        Test ValueError on passing dataframe not indexed by date.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm()
            data = self.data.reset_index()
            cla.allocate(asset_prices=data, solution='cla_turning_points', asset_names=self.data.columns)

    def test_value_error_for_unknown_returns(self):
        """
        Test ValueError on passing unknown returns string.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm(calculate_expected_returns="unknown_returns")
            cla.allocate(asset_prices=self.data, solution='cla_turning_points', asset_names=self.data.columns)

    def test_resampling_asset_prices(self):
        """
        Test resampling of asset prices.
        """

        cla = CriticalLineAlgorithm()
        cla.allocate(asset_prices=self.data, resample_by='B', solution='min_volatility', asset_names=self.data.columns)
        weights = cla.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_all_inputs_none(self):
        """
        Test allocation when all inputs are None.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm()
            cla.allocate(asset_names=self.data.columns)

    def test_cla_with_input_as_returns_and_covariance(self):
        # pylint: disable=invalid-name
        """
        Test CLA when we pass expected returns and covariance matrix as input.
        """

        cla = CriticalLineAlgorithm()
        expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data)
        covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data).cov()
        cla.allocate(covariance_matrix=covariance,
                     expected_asset_returns=expected_returns,
                     asset_names=self.data.columns)
        weights = cla.weights.values
        weights[weights <= 1e-15] = 0  # Convert very very small numbers to 0
        for turning_point in weights:
            assert (turning_point >= 0).all()
            assert len(turning_point) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(turning_point), 1)

    def test_no_asset_names(self):
        """
        Test CLA when not supplying a list of asset names.
        """

        cla = CriticalLineAlgorithm()
        cla.allocate(asset_prices=self.data)
        weights = cla.weights.values[0]
        assert (weights >= 0).all()
        assert len(weights) == self.data.shape[1]
        np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_valuerror_with_no_asset_names(self):
        """
        Test ValueError when not supplying a list of asset names and no other input.
        """

        with self.assertRaises(ValueError):
            cla = CriticalLineAlgorithm()
            expected_returns = ReturnsEstimators().calculate_mean_historical_returns(asset_prices=self.data,
                                                                                     resample_by='W')
            covariance = ReturnsEstimators().calculate_returns(asset_prices=self.data, resample_by='W').cov()
            cla.allocate(expected_asset_returns=expected_returns, covariance_matrix=covariance)



"""
Tests the unit functions in ch10_snippets.py for calculating bet size.
"""

import unittest
import warnings
import datetime as dt
import numpy as np
import pandas as pd

from scipy.stats import norm

from famlafl.bet_sizing.ch10_snippets import get_signal, avg_active_signals, mp_avg_active_signals, discrete_signal
from famlafl.bet_sizing.ch10_snippets import (bet_size_sigmoid, bet_size_power, bet_size, get_target_pos_sigmoid, get_target_pos_power, get_target_pos,
                                               inv_price_sigmoid, inv_price_power, inv_price, limit_price_sigmoid, limit_price_power, limit_price,
                                               get_w_sigmoid, get_w_power, get_w)


class TestCh10Snippets(unittest.TestCase):
    """
    Tests the following functions in ch10_snippets.py:
    - get_signal
    - avg_active_signals
    - mp_avg_active_signals
    - discrete_signal
    """

    def setUp(self):
        """
        Sets up the data to be used for the following tests.
        """
        # ======================================================================
        # Unit test setup for get_signal, avg_active_signals,
        # mp_avg_active_signals, and discrete_signal.

        # Setup the array of values, length of 6 for testing.
        prob_arr = np.array([0.711, 0.898, 0.992, 0.595, 0.544, 0.775])
        side_arr = np.array([1, 1, -1, 1, -1, 1])
        dates = np.array([dt.datetime(2000, 1, 1) + i*dt.timedelta(days=1) for i in range(6)])

        shift_list = [0.5, 1, 2, 1.5, 0.8, 0.2]
        shift_dt = np.array([dt.timedelta(days=d) for d in shift_list])
        dates_shifted = dates + shift_dt

        # Calculate the test statistic and bet size.
        z_test = (prob_arr - 0.5) / (prob_arr*(1-prob_arr))**0.5
        m_signal = side_arr * (2 * norm.cdf(z_test) - 1)
        m_discrete = np.array([max(-1, min(1, m_i)) for m_i in np.round(m_signal/0.1, 0)*0.1])

        # Convert arrays to pd.Series for use as test arguments.
        self.prob = pd.Series(data=prob_arr, index=dates)
        self.side = pd.Series(data=side_arr, index=dates)
        self.t_1 = pd.Series(data=dates_shifted, index=dates)
        self.bet_size = pd.Series(data=m_signal, index=dates)
        self.bet_size_d = pd.Series(data=m_discrete, index=dates)

        # Convert pd.Series to pd.DataFrames for calculating correct results.
        self.events = pd.concat(objs=[self.t_1, self.prob, self.side], axis=1)
        self.events = self.events.rename(columns={0: 't1', 1: 'prob', 2: 'side'})
        self.events_2 = self.events.copy()
        self.events_2['signal'] = self.bet_size

        # Calculation of the average active bets.
        t_p = set(self.events_2['t1'].to_numpy())
        t_p = t_p.union(self.events_2.index.to_numpy())
        t_p = list(t_p)
        t_p.sort()
        self.t_pnts = t_p.copy()
        avg_list = []
        for t_i in t_p:
            avg_list.append(self.events_2[(self.events_2.index <= t_i) & (self.events_2.t1 > t_i)]['signal'].mean())
        self.avg_active = pd.Series(data=np.array(avg_list), index=t_p).fillna(0)

    def test_get_signal(self):
        """
        Tests calculating the bet size from probability.
        """
        # Test get_signal when supplying a value to argument 'pred'.
        test_bet_size_1 = get_signal(prob=self.prob, num_classes=2, pred=self.side)
        self.assertEqual(self.bet_size.equals(test_bet_size_1), True)

        # Test get_signal when no value provided for 'pred'.
        test_bet_size_2 = get_signal(prob=self.prob, num_classes=2)
        self.assertEqual(self.bet_size.abs().equals(test_bet_size_2), True)

        # Test for prob.shape[0] == 0.
        df_empty = pd.DataFrame({'a': []})
        return_val = get_signal(df_empty, 2)
        self.assertIsInstance(return_val, pd.Series)
        self.assertEqual(0, len(return_val))

    def test_avg_active_signals(self):
        """
        Tests the avg_active_signals function. Also implicitly tests the
        molecular multiprocessing function mp_avg_active_signals.
        """
        test_avg_active = avg_active_signals(self.events_2)
        self.assertEqual(self.avg_active.equals(test_avg_active), True)

    def test_mp_avg_active_signals(self):
        """
        An explicit test of the mp_avg_active_signals subroutine.
        """
        test_mp_avg_active = mp_avg_active_signals(self.events_2, self.t_pnts)
        self.assertEqual(self.avg_active.equals(test_mp_avg_active), True)

    def test_discrete_signal(self):
        """
        Tests the discrete_signal function.
        """
        test_bet_discrete = discrete_signal(signal0=self.bet_size, step_size=0.1)
        self.assertEqual(self.bet_size_d.equals(test_bet_discrete), True)


# =====================================================================================================================
# Test cases for functions used in calculating dynamic bet size.
class TestBetSize(unittest.TestCase):
    """
    Test case for bet_size, bet_size_sigmoid, and bet_size_power.
    """
    def test_bet_size_sigmoid(self):
        """
        Tests successful execution of 'bet_size_sigmoid'.
        """
        x_div, w_param = 15, 7.5
        m_test = x_div / np.sqrt(w_param + x_div*x_div)
        self.assertAlmostEqual(m_test, bet_size_sigmoid(w_param, x_div), 7)

    def test_bet_size_power(self):
        """
        Tests successful execution of 'bet_size_power'.
        """
        x_div, w_param = 0.4, 1.5
        m_test = np.sign(x_div) * abs(x_div)**w_param
        self.assertAlmostEqual(m_test, bet_size_power(w_param, x_div), 7)

    def test_bet_size_power_value_error(self):
        """
        Tests successful raising of ValueError in 'bet_size_power'.
        """
        self.assertRaises(ValueError, bet_size_power, 2, 1.5)

    def test_bet_size_power_return_zero(self):
        """
        Tests that the function returns zero if the price divergence provided is zero.
        """
        self.assertEqual(0.0, bet_size_power(2, 0.0))

    def test_bet_size(self):
        """
        Test excution in all function modes of 'bet_size'.
        """
        x_div_sig, w_param_sig = 25, 3.5
        m_test_sig = x_div_sig / np.sqrt(w_param_sig + x_div_sig*x_div_sig)
        self.assertAlmostEqual(m_test_sig, bet_size(w_param_sig, x_div_sig, 'sigmoid'), 7)
        x_div_pow, w_param_pow = 0.7, 2.1
        m_test_pow = np.sign(x_div_pow) * abs(x_div_pow)**w_param_pow
        self.assertAlmostEqual(m_test_pow, bet_size(w_param_pow, x_div_pow, 'power'), 7)

    def test_bet_size_key_error(self):
        """
        Tests for the KeyError in the event that an invalid function is provided to 'func'.
        """
        self.assertRaises(KeyError, bet_size, 2, 3, 'NotAFunction')


class TestGetTPos(unittest.TestCase):
    """
    Test case for get_target_pos, get_target_pos_sigmoid, and get_target_pos_power.
    """
    def test_get_target_pos_sigmoid(self):
        """
        Tests successful execution of 'get_target_pos_sigmoid'.
        """
        f_i, m_p = 34.6, 21.9
        x_div = f_i - m_p
        w_param = 2.5
        max_pos = 200
        pos_test = int(max_pos*(x_div / np.sqrt(w_param + x_div*x_div)))
        self.assertAlmostEqual(pos_test, get_target_pos_sigmoid(w_param, f_i, m_p, max_pos), 7)

    def test_get_target_pos_power(self):
        """
        Tests successful execution of 'get_target_pos_power'.
        """
        f_i, m_p = 34.6, 34.1
        x_div = f_i - m_p
        w_param = 2.1
        max_pos = 100
        pos_test = int(max_pos*(np.sign(x_div) * abs(x_div)**w_param))
        self.assertAlmostEqual(pos_test, get_target_pos_power(w_param, f_i, m_p, max_pos), 7)

    def test_get_target_pos(self):
        """
        Tests successful execution in 'sigmoid' and 'power' function variants of 'get_target_pos'.
        """
        f_i_sig, m_p_sig = 31.6, 22.9
        x_div_sig = f_i_sig - m_p_sig
        w_param_sig = 2.6
        max_pos_sig = 220
        pos_test_sig = int(max_pos_sig*(x_div_sig / np.sqrt(w_param_sig + x_div_sig*x_div_sig)))
        self.assertAlmostEqual(pos_test_sig, get_target_pos(w_param_sig, f_i_sig, m_p_sig, max_pos_sig, 'sigmoid'), 7)
        f_i_pow, m_p_pow = 34.8, 34.1
        x_div_pow = f_i_pow - m_p_pow
        w_param_pow = 2.9
        max_pos_pow = 175
        pos_test_pow = int(max_pos_pow*(np.sign(x_div_pow) * abs(x_div_pow)**w_param_pow))
        self.assertAlmostEqual(pos_test_pow, get_target_pos(w_param_pow, f_i_pow, m_p_pow, max_pos_pow, 'power'), 7)

    def test_get_target_pos_key_error(self):
        """
        Tests for the KeyError in 'get_target_pos' in the case that an invalid value for 'func' is passed.
        """
        self.assertRaises(KeyError, get_target_pos, 1, 2, 1, 5, 'NotAFunction')


class TestInvPrice(unittest.TestCase):
    """
    Tests for functions 'inv_price', 'inv_price_sigmoid', and 'inv_price_power'.
    """
    def test_inv_price_sigmoid(self):
        """
        Test for the successful execution of 'inv_price_sigmoid'.
        """
        f_i_sig, w_sig, m_sig = 35.19, 9.32, 0.72
        inv_p_sig = f_i_sig - m_sig * np.sqrt(w_sig/(1-m_sig*m_sig))
        self.assertAlmostEqual(inv_p_sig, inv_price_sigmoid(f_i_sig, w_sig, m_sig), 7)

    def test_inv_price_power(self):
        """
        Test for the successful execution of 'inv_price_sigmoid'.
        """
        f_i_pow, w_pow, m_pow = 35.19, 3.32, 0.72
        inv_p_pow = f_i_pow - np.sign(m_pow) * abs(m_pow)**(1/w_pow)
        self.assertAlmostEqual(inv_p_pow, inv_price_power(f_i_pow, w_pow, m_pow), 7)
        self.assertEqual(f_i_pow, inv_price_power(f_i_pow, w_pow, 0.0))

    def test_inv_price(self):
        """
        Test for successful execution of 'inv_price' function under different function options.
        """
        f_i_sig, w_sig, m_sig = 87.19, 7.34, 0.82
        inv_p_sig = f_i_sig - m_sig * np.sqrt(w_sig/(1-m_sig*m_sig))
        self.assertAlmostEqual(inv_p_sig, inv_price(f_i_sig, w_sig, m_sig, 'sigmoid'), 7)
        f_i_pow, w_pow, m_pow = 129.19, 4.02, 0.81
        inv_p_pow = f_i_pow - np.sign(m_pow) * abs(m_pow)**(1/w_pow)
        self.assertAlmostEqual(inv_p_pow, inv_price(f_i_pow, w_pow, m_pow, 'power'), 7)

    def test_inv_price_key_error(self):
        """
        Test for successful raising of KeyError in response to invalid choice of 'func' argument.
        """
        self.assertRaises(KeyError, inv_price, 12, 1.5, 0.7, 'NotAFunction')


class TestLimitPrice(unittest.TestCase):
    """
    Tests the functions 'limit_price_sigmoid', 'limit_price_power', and 'limit_price'.
    """
    def test_limit_price_sigmoid(self):
        """
        Test successful execution of 'limit_price_sigmoid' function.
        """
        t_pos_sig, pos_sig, f_sig, w_sig, max_pos_sig = 124, 112, 165.50, 8.44, 150
        sum_inv_price_sig = sum([inv_price_sigmoid(f_sig, w_sig, j/float(max_pos_sig))
                                 for j in range(abs(pos_sig+np.sign(t_pos_sig-pos_sig)), abs(t_pos_sig+1))])
        limit_p_sig = (1/abs(t_pos_sig-pos_sig)) * sum_inv_price_sig
        self.assertAlmostEqual(limit_p_sig, limit_price_sigmoid(t_pos_sig, pos_sig, f_sig, w_sig, max_pos_sig), 7)

    def test_limit_price_sigmoid_return_nan(self):
        """
        Tests for the successful return of np.nan in the case that the target position is the same as the
        current position.
        """
        self.assertTrue(np.isnan(limit_price_sigmoid(1, 1, 123, 21, 234)))

    def test_limit_price_power(self):
        """
        Test successful execution of 'limit_price_power' function.
        """
        t_pos_pow, pos_pow, f_pow, w_pow, max_pos_pow = 101, 95, 195.70, 3.44, 130
        sum_inv_price_pow = sum([inv_price_power(f_pow, w_pow, j/float(max_pos_pow))
                                 for j in range(abs(pos_pow+np.sign(t_pos_pow-pos_pow)), abs(t_pos_pow+1))])
        limit_p_pow = (1/abs(t_pos_pow-pos_pow)) * sum_inv_price_pow
        self.assertAlmostEqual(limit_p_pow, limit_price_power(t_pos_pow, pos_pow, f_pow, w_pow, max_pos_pow), 7)

    def test_limit_price_key_error(self):
        """
        Tests raising of the KeyError due to invalid choice of 'func' argument.
        """
        self.assertRaises(KeyError, limit_price, 231, 221, 110, 3.4, 250, 'NotAFunction')


class TestGetW(unittest.TestCase):
    """
    Tests the functions 'get_w_sigmoid', 'get_w_power', and 'get_w'.
    """
    def test_get_w_sigmoid(self):
        """
        Tests successful execution of 'get_w_sigmoid' function.
        """
        x_sig, m_sig = 24.2, 0.98
        w_sig = x_sig**2 * (m_sig**-2 - 1)
        self.assertAlmostEqual(w_sig, get_w_sigmoid(x_sig, m_sig), 7)

    def test_get_w_power(self):
        """
        Tests successful execution of 'get_w_power' function.
        """
        x_pow, m_pow = 0.9, 0.76
        w_pow = np.log(m_pow/np.sign(x_pow)) / np.log(abs(x_pow))
        self.assertAlmostEqual(w_pow, get_w_power(x_pow, m_pow), 7)

    def test_get_w_power_value_error(self):
        """
        Tests that a ValueError is raised if the price divergence 'x' is not between -1 and 1, inclusive.
        """
        self.assertRaises(ValueError, get_w_power, 1.2, 0.8)

    def test_get_w_power_warning(self):
        """
        Tests that a UserWarning is raised if 'w' is calcualted to be less than zero, and returns a zero.
        """
        with warnings.catch_warnings(record=True) as warn_catch:
            warnings.simplefilter("always")
            w_param = get_w_power(0.1, 2)
            self.assertTrue(len(warn_catch) == 1)
            self.assertEqual(0, w_param)
            self.assertTrue('User' in str(warn_catch[0].category))

    def test_get_w_key_error(self):
        """
        Tests that a KeyError is raised if an invalid function is passed to argument 'func'.
        """
        self.assertRaises(KeyError, get_w, 0.6, 0.9, 'NotAFunction')



"""
A base class for the various bar types. Includes the logic shared between classes, to minimise the amount of
duplicated code.
"""

from abc import ABC, abstractmethod
from typing import Tuple, Union, Generator, Iterable, Optional

import numpy as np
import pandas as pd

from famlafl.util.fast_ewma import ewma


def _crop_data_frame_in_batches(df: pd.DataFrame, chunksize: int) -> list:
    # pylint: disable=invalid-name
    """
    Splits df into chunks of chunksize

    :param df: (pd.DataFrame) Dataframe to split
    :param chunksize: (int) Number of rows in chunk
    :return: (list) Chunks (pd.DataFrames)
    """
    generator_object = []
    for _, chunk in df.groupby(np.arange(len(df)) // chunksize):
        generator_object.append(chunk)
    return generator_object

# pylint: disable=too-many-instance-attributes


class BaseBars(ABC):
    """
    Abstract base class which contains the structure which is shared between the various standard and information
    driven bars. There are some methods contained in here that would only be applicable to information bars but
    they are included here so as to avoid a complicated nested class structure.
    """

    def __init__(self, metric: str, batch_size: int = 2e7):
        """
        Constructor

        :param metric: (str) Type of imbalance bar to create. Example: dollar_imbalance.
        :param batch_size: (int) Number of rows to read in from the csv, per batch.
        """

        # Base properties
        self.metric = metric
        self.batch_size = batch_size
        self.prev_tick_rule = 0

        # Cache properties
        self.open_price, self.prev_price, self.close_price = None, None, None
        self.high_price, self.low_price = -np.inf, np.inf
        self.cum_statistics = {'cum_ticks': 0, 'cum_dollar_value': 0, 'cum_volume': 0, 'cum_buy_volume': 0}
        self.tick_num = 0  # Tick number when bar was formed

        # Batch_run properties
        self.flag = False  # The first flag is false since the first batch doesn't use the cache


    def batch_run(self, file_path_or_df: Union[str, Iterable[str], pd.DataFrame], verbose: bool = True, to_csv: bool = False,
                  output_path: Optional[str] = None) -> Union[pd.DataFrame, None]:
        """
        Reads csv file(s) or pd.DataFrame in batches and then constructs the financial data structure in the form of a DataFrame.
        The csv file or DataFrame must have only 3 columns: date_time, price, & volume.

        :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing
                                raw tick data  in the format[date_time, price, volume]
        :param verbose: (bool) Flag whether to print message on each processed batch or not
        :param to_csv: (bool) Flag for writing the results of bars generation to local csv file, or to in-memory DataFrame
        :param output_path: (bool) Path to results file, if to_csv = True

        :return: (pd.DataFrame or None) Financial data structure
        """

        if to_csv is True:
            header = True  # if to_csv is True, header should written on the first batch only
            open(output_path, 'w').close()  # clean output csv file

        if verbose:  # pragma: no cover
            print('Reading data in batches:')

        # Read csv in batches
        count = 0
        final_bars = []
        cols = ['date_time', 'tick_num', 'open', 'high', 'low', 'close', 'volume', 'cum_buy_volume', 'cum_ticks',
                'cum_dollar_value']
        for batch in self._batch_iterator(file_path_or_df):
            if verbose:  # pragma: no cover
                print('Batch number:', count)

            list_bars = self.run(data=batch)

            if to_csv is True:
                pd.DataFrame(list_bars, columns=cols).to_csv(output_path, header=header, index=False, mode='a')
                header = False
            else:
                # Append to bars list
                final_bars += list_bars
            count += 1

        if verbose:  # pragma: no cover
            print('Returning bars \n')

        # Return a DataFrame
        if final_bars:
            bars_df = pd.DataFrame(final_bars, columns=cols)
            return bars_df

        # Processed DataFrame is stored in .csv file, return None
        return None

    def _batch_iterator(self, file_path_or_df: Union[str, Iterable[str], pd.DataFrame]) -> Generator[pd.DataFrame, None, None]:
        """
        :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame
                                containing raw tick data in the format[date_time, price, volume]
        """
        if isinstance(file_path_or_df, (list, tuple)):
            # Assert format of all files
            for file_path in file_path_or_df:
                self._read_first_row(file_path)
            for file_path in file_path_or_df:
                for batch in pd.read_csv(file_path, chunksize=self.batch_size, parse_dates=[0]):
                    yield batch

        elif isinstance(file_path_or_df, str):
            self._read_first_row(file_path_or_df)
            for batch in pd.read_csv(file_path_or_df, chunksize=self.batch_size, parse_dates=[0]):
                yield batch

        elif isinstance(file_path_or_df, pd.DataFrame):
            for batch in _crop_data_frame_in_batches(file_path_or_df, self.batch_size):
                yield batch

        else:
            raise ValueError('file_path_or_df is neither string(path to a csv file), iterable of strings, nor pd.DataFrame')

    def _read_first_row(self, file_path: str):
        """
        :param file_path: (str) Path to the csv file containing raw tick data in the format[date_time, price, volume]
        """
        # Read in the first row & assert format
        first_row = pd.read_csv(file_path, nrows=1)
        self._assert_csv(first_row)

    def run(self, data: Union[list, tuple, pd.DataFrame]) -> list:
        """
        Reads a List, Tuple, or Dataframe and then constructs the financial data structure in the form of a list.
        The List, Tuple, or DataFrame must have only 3 attrs: date_time, price, & volume.

        :param data: (list, tuple, or pd.DataFrame) Dict or ndarray containing raw tick data in the format[date_time, price, volume]

        :return: (list) Financial data structure
        """

        if isinstance(data, (list, tuple)):
            values = data

        elif isinstance(data, pd.DataFrame):
            values = data.values

        else:
            raise ValueError('data is neither list nor tuple nor pd.DataFrame')

        list_bars = self._extract_bars(data=values)

        # Set flag to True: notify function to use cache
        self.flag = True

        return list_bars

    @abstractmethod
    def _extract_bars(self, data: pd.DataFrame) -> list:
        """
        This method is required by all the bar types and is used to create the desired bars.

        :param data: (pd.DataFrame) Contains 3 columns - date_time, price, and volume.
        :return: (list) Bars built using the current batch.
        """

    @abstractmethod
    def _reset_cache(self):
        """
        This method is required by all the bar types. It describes how cache should be reset
        when new bar is sampled.
        """

    @staticmethod
    def _assert_csv(test_batch: pd.DataFrame):
        """
        Tests that the csv file read has the format: date_time, price, and volume.
        If not then the user needs to create such a file. This format is in place to remove any unwanted overhead.

        :param test_batch: (pd.DataFrame) The first row of the dataset.
        """
        assert test_batch.shape[1] == 3, 'Must have only 3 columns in csv: date_time, price, & volume.'
        assert isinstance(test_batch.iloc[0, 1], float), 'price column in csv not float.'
        assert not isinstance(test_batch.iloc[0, 2], str), 'volume column in csv not int or float.'

        try:
            pd.to_datetime(test_batch.iloc[0, 0])
        except ValueError:
            raise ValueError('csv file, column 0, not a date time format:',
                             test_batch.iloc[0, 0])

    def _update_high_low(self, price: float) -> Union[float, float]:
        """
        Update the high and low prices using the current price.

        :param price: (float) Current price
        :return: (tuple) Updated high and low prices
        """
        if price > self.high_price:
            high_price = price
        else:
            high_price = self.high_price

        if price < self.low_price:
            low_price = price
        else:
            low_price = self.low_price

        return high_price, low_price

    def _create_bars(self, date_time: str, price: float, high_price: float, low_price: float, list_bars: list) -> None:
        """
        Given the inputs, construct a bar which has the following fields: date_time, open, high, low, close, volume,
        cum_buy_volume, cum_ticks, cum_dollar_value.
        These bars are appended to list_bars, which is later used to construct the final bars DataFrame.

        :param date_time: (str) Timestamp of the bar
        :param price: (float) The current price
        :param high_price: (float) Highest price in the period
        :param low_price: (float) Lowest price in the period
        :param list_bars: (list) List to which we append the bars
        """
        # Create bars
        open_price = self.open_price
        high_price = max(high_price, open_price)
        low_price = min(low_price, open_price)
        close_price = price
        volume = self.cum_statistics['cum_volume']
        cum_buy_volume = self.cum_statistics['cum_buy_volume']
        cum_ticks = self.cum_statistics['cum_ticks']
        cum_dollar_value = self.cum_statistics['cum_dollar_value']

        # Create a new row as a list
        new_row = [date_time, self.tick_num, open_price, high_price, low_price,
                  close_price, volume, cum_buy_volume, cum_ticks, cum_dollar_value]
        
        # Append the new row to list_bars
        list_bars.append(new_row)

    def _apply_tick_rule(self, price: float) -> int:
        """
        Applies the tick rule as defined on page 29 of Advances in Financial Machine Learning.

        :param price: (float) Price at time t
        :return: (int) The signed tick
        """
        if self.prev_price is not None:
            tick_diff = price - self.prev_price
        else:
            tick_diff = 0

        if tick_diff != 0:
            signed_tick = np.sign(tick_diff)
            self.prev_tick_rule = signed_tick
        else:
            signed_tick = self.prev_tick_rule

        self.prev_price = price  # Update previous price used for tick rule calculations
        return signed_tick

    def _get_imbalance(self, price: float, signed_tick: int, volume: float) -> float:
        """
        Advances in Financial Machine Learning, page 29.

        Get the imbalance at a point in time, denoted as Theta_t

        :param price: (float) Price at t
        :param signed_tick: (int) signed tick, using the tick rule
        :param volume: (float) Volume traded at t
        :return: (float) Imbalance at time t
        """
        if self.metric == 'tick_imbalance' or self.metric == 'tick_run':
            imbalance = signed_tick
        elif self.metric == 'dollar_imbalance' or self.metric == 'dollar_run':
            imbalance = signed_tick * volume * price
        elif self.metric == 'volume_imbalance' or self.metric == 'volume_run':
            imbalance = signed_tick * volume
        else:
            raise ValueError('Unknown imbalance metric, possible values are tick/dollar/volume imbalance/run')
        return imbalance


class BaseImbalanceBars(BaseBars):
    """
    Base class for Imbalance Bars (EMA and Const) which implements imbalance bars calculation logic
    """

    def __init__(self, metric: str, batch_size: int,
                 expected_imbalance_window: int, exp_num_ticks_init: int,
                 analyse_thresholds: bool):
        """
        Constructor

        :param metric: (str) Type of imbalance bar to create. Example: dollar_imbalance.
        :param batch_size: (int) Number of rows to read in from the csv, per batch.
        :param expected_imbalance_window: (int) Window used to estimate expected imbalance from previous trades
        :param exp_num_ticks_init: (int) Initial estimate for expected number of ticks in bar.
                                         For Const Imbalance Bars expected number of ticks equals expected number of ticks init
        :param analyse_thresholds: (bool) Flag to return thresholds values (theta, exp_num_ticks, exp_imbalance) in a
                                          form of Pandas DataFrame
        """
        BaseBars.__init__(self, metric, batch_size)

        self.expected_imbalance_window = expected_imbalance_window

        self.thresholds = {'cum_theta': 0, 'expected_imbalance': np.nan, 'exp_num_ticks': exp_num_ticks_init}

        # Previous bars number of ticks and previous tick imbalances
        self.imbalance_tick_statistics = {'num_ticks_bar': [], 'imbalance_array': []}

        if analyse_thresholds is True:
            # Array of dicts: {'timestamp': value, 'cum_theta': value, 'exp_num_ticks': value, 'exp_imbalance': value}
            self.bars_thresholds = []
        else:
            self.bars_thresholds = None

    def _reset_cache(self):
        """
        Implementation of abstract method _reset_cache for imbalance bars
        """
        self.open_price = None
        self.high_price, self.low_price = -np.inf, np.inf
        self.cum_statistics = {'cum_ticks': 0, 'cum_dollar_value': 0, 'cum_volume': 0, 'cum_buy_volume': 0}
        self.thresholds['cum_theta'] = 0

    def _extract_bars(self, data: Tuple[dict, pd.DataFrame]) -> list:
        """
        For loop which compiles the various imbalance bars: dollar, volume, or tick.

        :param data: (pd.DataFrame) Contains 3 columns - date_time, price, and volume.
        :return: (list) Bars built using the current batch.
        """

        # Iterate over rows
        list_bars = []
        for row in data:
            # Set variables
            date_time = row[0]
            self.tick_num += 1
            price = float(row[1])
            volume = row[2]
            dollar_value = price * volume
            signed_tick = self._apply_tick_rule(price)

            if self.open_price is None:
                self.open_price = price

            # Update high low prices
            self.high_price, self.low_price = self._update_high_low(price)

            # Bar statistics calculations
            self.cum_statistics['cum_ticks'] += 1
            self.cum_statistics['cum_dollar_value'] += dollar_value
            self.cum_statistics['cum_volume'] += volume
            if signed_tick == 1:
                self.cum_statistics['cum_buy_volume'] += volume

            # Imbalance calculations
            imbalance = self._get_imbalance(price, signed_tick, volume)
            self.imbalance_tick_statistics['imbalance_array'].append(imbalance)
            self.thresholds['cum_theta'] += imbalance

            # Get expected imbalance for the first time, when num_ticks_init passed
            if not list_bars and np.isnan(self.thresholds['expected_imbalance']):
                self.thresholds['expected_imbalance'] = self._get_expected_imbalance(
                    self.expected_imbalance_window)

            if self.bars_thresholds is not None:
                self.thresholds['timestamp'] = date_time
                self.bars_thresholds.append(dict(self.thresholds))

            # Check expression for possible bar generation
            if (np.abs(self.thresholds['cum_theta']) > self.thresholds['exp_num_ticks'] * np.abs(
                    self.thresholds['expected_imbalance']) if ~np.isnan(self.thresholds['expected_imbalance']) else False):
                self._create_bars(date_time, price,
                                  self.high_price, self.low_price, list_bars)

                self.imbalance_tick_statistics['num_ticks_bar'].append(self.cum_statistics['cum_ticks'])
                # Expected number of ticks based on formed bars
                self.thresholds['exp_num_ticks'] = self._get_exp_num_ticks()
                # Get expected imbalance
                self.thresholds['expected_imbalance'] = self._get_expected_imbalance(
                    self.expected_imbalance_window)
                # Reset counters
                self._reset_cache()

        return list_bars

    def _get_expected_imbalance(self, window: int):
        """
        Calculate the expected imbalance: 2P[b_t=1]-1, using a EWMA, pg 29
        :param window: (int) EWMA window for calculation
        :return: expected_imbalance: (np.ndarray) 2P[b_t=1]-1, approximated using a EWMA
        """
        if len(self.imbalance_tick_statistics['imbalance_array']) < self.thresholds['exp_num_ticks']:
            # Waiting for array to fill for ewma
            ewma_window = np.nan
        else:
            # ewma window can be either the window specified in a function call
            # or it is len of imbalance_array if window > len(imbalance_array)
            ewma_window = int(min(len(self.imbalance_tick_statistics['imbalance_array']), window))

        if np.isnan(ewma_window):
            # return nan, wait until len(self.imbalance_array) >= self.exp_num_ticks_init
            expected_imbalance = np.nan
        else:
            expected_imbalance = ewma(
                np.array(self.imbalance_tick_statistics['imbalance_array'][-ewma_window:], dtype=float),
                window=ewma_window)[-1]

        return expected_imbalance

    @abstractmethod
    def _get_exp_num_ticks(self):
        """
        Abstract method which updates expected number of ticks when new run bar is formed
        """


# pylint: disable=too-many-instance-attributes
class BaseRunBars(BaseBars):
    """
    Base class for Run Bars (EMA and Const) which implements run bars calculation logic
    """

    def __init__(self, metric: str, batch_size: int, num_prev_bars: int,
                 expected_imbalance_window: int,
                 exp_num_ticks_init: int, analyse_thresholds: bool):
        """
        Constructor

        :param metric: (str) Type of imbalance bar to create. Example: dollar_imbalance.
        :param batch_size: (int) Number of rows to read in from the csv, per batch.
        :param expected_imbalance_window: (int) Window used to estimate expected imbalance from previous trades
        :param exp_num_ticks_init: (int) Initial estimate for expected number of ticks in bar.
                                         For Const Imbalance Bars expected number of ticks equals expected number of ticks init
        :param analyse_thresholds: (bool) Flag to return thresholds values (thetas, exp_num_ticks, exp_runs) in Pandas DataFrame
        """
        BaseBars.__init__(self, metric, batch_size)

        self.num_prev_bars = num_prev_bars
        self.expected_imbalance_window = expected_imbalance_window

        self.thresholds = {'cum_theta_buy': 0, 'cum_theta_sell': 0, 'exp_imbalance_buy': np.nan,
                           'exp_imbalance_sell': np.nan, 'exp_num_ticks': exp_num_ticks_init,
                           'exp_buy_ticks_proportion': np.nan, 'buy_ticks_num': 0}

        # Previous bars number of ticks and previous tick imbalances
        self.imbalance_tick_statistics = {'num_ticks_bar': [], 'imbalance_array_buy': [], 'imbalance_array_sell': [],
                                          'buy_ticks_proportion': []}

        if analyse_thresholds:
            # Array of dicts: {'timestamp': value, 'cum_theta': value, 'exp_num_ticks': value, 'exp_imbalance': value}
            self.bars_thresholds = []
        else:
            self.bars_thresholds = None

        self.warm_up_flag = False

    def _reset_cache(self):
        """
        Implementation of abstract method _reset_cache for imbalance bars
        """
        self.open_price = None
        self.high_price, self.low_price = -np.inf, np.inf
        self.cum_statistics = {'cum_ticks': 0, 'cum_dollar_value': 0, 'cum_volume': 0, 'cum_buy_volume': 0}
        self.thresholds['cum_theta_buy'], self.thresholds['cum_theta_sell'], self.thresholds['buy_ticks_num'] = 0, 0, 0

    def _extract_bars(self, data: Tuple[list, np.ndarray]) -> list:
        """
        For loop which compiles the various run bars: dollar, volume, or tick.

        :param data: (list or np.ndarray) Contains 3 columns - date_time, price, and volume.
        :return: (list) of bars built using the current batch.
        """

        # Iterate over rows
        list_bars = []
        for row in data:
            # Set variables
            date_time = row[0]
            self.tick_num += 1
            price = float(row[1])
            volume = row[2]
            dollar_value = price * volume
            signed_tick = self._apply_tick_rule(price)

            if self.open_price is None:
                self.open_price = price

            # Update high low prices
            self.high_price, self.low_price = self._update_high_low(price)

            # Bar statistics calculations
            self.cum_statistics['cum_ticks'] += 1
            self.cum_statistics['cum_dollar_value'] += dollar_value
            self.cum_statistics['cum_volume'] += volume
            if signed_tick == 1:
                self.cum_statistics['cum_buy_volume'] += volume

            # Imbalance calculations
            imbalance = self._get_imbalance(price, signed_tick, volume)

            if imbalance > 0:
                self.imbalance_tick_statistics['imbalance_array_buy'].append(imbalance)
                self.thresholds['cum_theta_buy'] += imbalance
                self.thresholds['buy_ticks_num'] += 1
            elif imbalance < 0:
                self.imbalance_tick_statistics['imbalance_array_sell'].append(abs(imbalance))
                self.thresholds['cum_theta_sell'] += abs(imbalance)

            self.warm_up_flag = np.isnan([self.thresholds['exp_imbalance_buy'], self.thresholds[
                'exp_imbalance_sell']]).any()  # Flag indicating that one of imbalances is not counted (warm-up)

            # Get expected imbalance for the first time, when num_ticks_init passed
            if not list_bars and self.warm_up_flag:
                self.thresholds['exp_imbalance_buy'] = self._get_expected_imbalance(
                    self.imbalance_tick_statistics['imbalance_array_buy'], self.expected_imbalance_window, warm_up=True)
                self.thresholds['exp_imbalance_sell'] = self._get_expected_imbalance(
                    self.imbalance_tick_statistics['imbalance_array_sell'], self.expected_imbalance_window,
                    warm_up=True)

                if bool(np.isnan([self.thresholds['exp_imbalance_buy'],
                                  self.thresholds['exp_imbalance_sell']]).any()) is False:
                    self.thresholds['exp_buy_ticks_proportion'] = self.thresholds['buy_ticks_num'] / \
                                                                  self.cum_statistics[
                                                                      'cum_ticks']

            if self.bars_thresholds is not None:
                self.thresholds['timestamp'] = date_time
                self.bars_thresholds.append(dict(self.thresholds))

            # Check expression for possible bar generation
            max_proportion = max(
                self.thresholds['exp_imbalance_buy'] * self.thresholds['exp_buy_ticks_proportion'],
                self.thresholds['exp_imbalance_sell'] * (1 - self.thresholds['exp_buy_ticks_proportion']))

            # Check expression for possible bar generation
            max_theta = max(self.thresholds['cum_theta_buy'], self.thresholds['cum_theta_sell'])
            if max_theta > self.thresholds['exp_num_ticks'] * max_proportion and not np.isnan(max_proportion):
                self._create_bars(date_time, price, self.high_price, self.low_price, list_bars)

                self.imbalance_tick_statistics['num_ticks_bar'].append(self.cum_statistics['cum_ticks'])
                self.imbalance_tick_statistics['buy_ticks_proportion'].append(
                    self.thresholds['buy_ticks_num'] / self.cum_statistics['cum_ticks'])

                # Expected number of ticks based on formed bars
                self.thresholds['exp_num_ticks'] = self._get_exp_num_ticks()

                # Expected buy ticks proportion based on formed bars
                exp_buy_ticks_proportion = ewma(
                    np.array(self.imbalance_tick_statistics['buy_ticks_proportion'][-self.num_prev_bars:], dtype=float),
                    self.num_prev_bars)[-1]
                self.thresholds['exp_buy_ticks_proportion'] = exp_buy_ticks_proportion

                # Get expected imbalance
                self.thresholds['exp_imbalance_buy'] = self._get_expected_imbalance(
                    self.imbalance_tick_statistics['imbalance_array_buy'], self.expected_imbalance_window)
                self.thresholds['exp_imbalance_sell'] = self._get_expected_imbalance(
                    self.imbalance_tick_statistics['imbalance_array_sell'], self.expected_imbalance_window)

                # Reset counters
                self._reset_cache()

        return list_bars

    def _get_expected_imbalance(self, array: list, window: int, warm_up: bool = False):
        """
        Advances in Financial Machine Learning, page 29.

        Calculates the expected imbalance: 2P[b_t=1]-1, using a EWMA.

        :param array: (list) of imbalances
        :param window: (int) EWMA window for calculation
        :parawm warm_up: (bool) flag of whether warm up period passed
        :return: expected_imbalance: (np.ndarray) 2P[b_t=1]-1, approximated using a EWMA
        """
        if len(array) < self.thresholds['exp_num_ticks'] and warm_up is True:
            # Waiting for array to fill for ewma
            ewma_window = np.nan
        else:
            # ewma window can be either the window specified in a function call
            # or it is len of imbalance_array if window > len(imbalance_array)
            ewma_window = int(min(len(array), window))

        if np.isnan(ewma_window):
            # return nan, wait until len(self.imbalance_array) >= self.exp_num_ticks_init
            expected_imbalance = np.nan
        else:
            expected_imbalance = ewma(
                np.array(array[-ewma_window:], dtype=float),
                window=ewma_window)[-1]

        return expected_imbalance

    @abstractmethod
    def _get_exp_num_ticks(self):
        """
        Abstract method which updates expected number of ticks when new imbalance bar is formed
        """



"""
Tests Best Stock.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection import BestStock


class TestBestStock(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Best Stock class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_best_stock_solution(self):
        """
        Tests that the weights sum to 1.
        """
        # Initialize BestStock.
        beststock = BestStock()
        # Allocates asset prices to BestStock.
        beststock.allocate(self.data)
        # Create np.array of all_weights.
        all_weights = np.array(beststock.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_best_performing(self):
        """
        Tests that returning weights indicate the best performing asset.
        """
        # Initialize BestStock.
        beststock1 = BestStock()
        # Allocates asset prices to BestStock.
        beststock1.allocate(self.data)
        # Best performing asset calculated by dividing the last row by the first row.
        price_diff = np.array(self.data.iloc[-1] / self.data.iloc[0])
        # Index of best stock through manual calculation.
        idx_price_diff = np.argmax(price_diff)
        # Weight returned by BestStock.
        beststock_weight = np.array(beststock1.all_weights)[0]
        # Index of the best stock through BestStock.
        idx_best_stock = np.argmax(beststock_weight)
        # The indices should be identical.
        np.testing.assert_equal(idx_best_stock, idx_price_diff)

    def test_number_of_nonzero(self):
        """
        Tests that the weights returned have only one value that is non-zero.
        """
        # Initialize BestStock.
        beststock = BestStock()
        # Allocates asset prices to BestStock.
        beststock.allocate(self.data)
        # Weight returned by BestStock.
        beststock_weight = np.array(beststock.all_weights)[0]
        # There should be only one non-zero value for the weights.
        np.testing.assert_equal(np.count_nonzero(beststock_weight), 1)



# pylint: disable=missing-module-docstring
import cvxpy as cp
from famlafl.online_portfolio_selection.base import OLPS


class BCRP(OLPS):
    """
    This class implements the Best Constant Rebalanced Portfolio strategy. It is reproduced with
    modification from the following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    Best Constant Rebalanced Portfolio rebalances to a set of weight that maximizes returns over a
    given time period. This strategy is implemented in hindsight and is not predictive.
    """

    def _first_weight(self, weights):
        """
        Returns the first weight of the given portfolio to be the Best Constant Rebalanced Portfolio
        in hindsight.

        :param weights: (np.array) Given weights by the user.
        :return: (np.array) Weights that maximize the returns.
        """
        # Use cp.SCS solver to speed up calculations.
        new_weights = self._optimize(self.relative_return, solver=cp.SCS)
        return new_weights



"""
Tests Best Constant Rebalanced Portfolio.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection import BCRP, CRP


class TestBestConstantRebalancedPortfolio(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    # pylint: disable=protected-access
    """
    Tests different functions of the Best Constant Rebalanced Portfolio class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_bcrp_solution(self):
        """
        Tests the calculation of best constant rebalanced portfolio weights.
        """
        # Initialize BCRP.
        bcrp = BCRP()
        # Allocates asset prices to BCRP.
        bcrp.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(bcrp.all_weights)
        # All weights for the strategy have to be the same.
        one_weight = all_weights[0]
        # iterate through all_weights to check that weights equal to the first weight.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            assert (weights == one_weight).all()
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_bcrp_returns(self):
        """
        Tests that BCRP returns are higher than other CRP's.
        """
        # Initialize BCRP
        bcrp1 = BCRP()
        # Allocates asset prices to BCRP.
        bcrp1.allocate(self.data, resample_by='M')
        # Get final returns for bcrp1.
        bcrp1_returns = np.array(bcrp1.portfolio_return)[-1]
        # Set an arbitray weight to test.
        weight = bcrp1._uniform_weight()
        # Initialize CRP.
        crp = CRP(weight)
        crp.allocate(self.data, resample_by='M')
        # Get final returns for CRP.
        crp_returns = np.array(crp.portfolio_return)[-1]
        # Check that CRP returns are lower than BCRP returns.
        np.testing.assert_array_less(crp_returns, bcrp1_returns)



# pylint: disable=missing-module-docstring
import numpy as np
from scipy.stats import norm
from famlafl.online_portfolio_selection.base import OLPS


class CWMR(OLPS):
    """
    This class implements the Confidence Weighted Mean Reversion strategy. It is reproduced with
    modification from the following paper:
    `Li, B., Hoi, S.C., Zhao, P. & Gopalkrishnan, V.. (2011). Confidence Weighted Mean Reversion
    Strategy for On-Line Portfolio Selection. Proceedings of the Fourteenth International
    Conference on Artificial Intelligence and Statistics, in PMLR 15:434-442.
    <https://ink.library.smu.edu.sg/cgi/viewcontent.cgi?article=3292&context=sis_research>`_

    Confidence Weighted Mean Reversion exploits both the popular mean reversion techniques and
    second-order information to model weights as a gaussian distribution.
    """

    def __init__(self, confidence, epsilon, method='var'):
        """
        Initializes Confidence Weighted Mean Reversion with the given confidence, epsilon, and method.

        :param confidence: (float) Confidence parameter with range of [0, 1]. CWMR is extremely
                                   sensitive to parameters, and it is unpredictable how the confidence
                                   value directly affects the results. Typically the extreme values of
                                   0 and 1 have the highest returns. A low value indicates a wider
                                   range of selection for weights.
        :param epsilon: (float) Mean reversion parameter with range of [0, 1]. CWMR is extremely
                                sensitive to parameters, and it is unpredictable how the epsilon value
                                directly affects the results. Typically the extreme values of 0 and
                                1 have the highest returns.
        :param method: (string) Variance update method. Choose 'var' for variance and 'sd' for
                                standard deviation. Both methods are dependant on the confidence and
                                epsilon parameters more than the method itself.
        """
        self.confidence = confidence
        self.theta = None
        self.epsilon = epsilon
        self.method = method
        self.sigma = None  # (np.array) Variance of the portfolio distribution.
        self.mu_dist = None  # (np.array) Mean of the portfolio distribution.
        super().__init__()

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                 'M' for Month. The inputs are based on pandas' resample method.
        """
        super(CWMR, self)._initialize(asset_prices, weights, resample_by)

        # Check that epsilon value is correct.
        if self.epsilon < 0 or self.epsilon > 1:
            raise ValueError("Epsilon values must be between 0 and 1.")

        # Check that the confidence value is correct.
        if self.confidence < 0 or self.confidence > 1:
            raise ValueError("Confidence values must be between 0 and 1.")

        # If confidence is valid, set theta value.
        self.theta = norm.ppf(self.confidence)

        # Check that the given method is correct.
        if self.method != 'var' and self.method != 'sd':
            raise ValueError("Method must be either 'var' or 'sd'.")

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) Predicted weights.
        """
        # Set current relative returns.
        curr_relative_return = self.relative_return[[time]]

        # Calculate dot product of relative returns and current weights.
        new_m = np.dot(curr_relative_return, self.weights)[0]

        # Calculate the variance of current relative returns.
        new_v = np.dot(np.dot(curr_relative_return, self.sigma), curr_relative_return.T)[0][0]

        # Calculate weighted variance.
        new_w = np.dot(np.dot(curr_relative_return, self.sigma), np.identity(self.number_of_assets))[0]

        # Calculate weighted average.
        mean_x = np.diag(self.sigma) / np.trace(self.sigma)

        # Calculate lambd.
        lambd = self._calculate_lambd(new_m, new_v, new_w, mean_x)

        # Update mu.
        self.mu_dist -= lambd * np.dot(curr_relative_return - mean_x, self.sigma).reshape((self.number_of_assets,))

        # Calculate new sigma.
        if self.method == 'sd':
            # Component for new variance calculation.
            sqrt_u = (-lambd * self.theta * new_v + np.sqrt(lambd ** 2 * self.theta ** 2 * new_v ** 2 + 4 * new_v)) / 2
            # Update variance.
            self.sigma = np.linalg.pinv(np.linalg.pinv(self.sigma) + lambd * self.theta / sqrt_u * np.diag(curr_relative_return) ** 2)

        if self.method == 'var':
            # Update variance.
            self.sigma = np.linalg.pinv(
                np.linalg.pinv(self.sigma) + 2 * lambd * self.theta * np.diag(curr_relative_return) ** 2)

        # Change to Positive Semi-Definite Matrix.
        self.sigma[self.sigma < 1e-5 * np.identity(self.number_of_assets)] = 1e-5

        # Normalize variance.
        self.sigma /= new_m * np.trace(self.sigma)

        # Simplex projection.
        self.mu_dist = self._simplex_projection(self.mu_dist)
        new_weights = self.mu_dist
        return new_weights

    def _first_weight(self, weights):
        """
        Returns the first weight of the given portfolio. If the first weight is not given,
        initialize weights to uniform weights.

        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :return: (np.array) First portfolio weight.
        """
        # Set sigma, the variance of the portfolio distribution.
        self.sigma = np.identity(self.number_of_assets) / (self.number_of_assets ** 2)

        # If no weights are given, return uniform weights.
        if weights is None:
            weights = self._uniform_weight()

        # Set mu, the mean of the portfolio distribution.
        self.mu_dist = weights
        return weights

    def _calculate_lambd(self, new_m, new_v, new_w, mean_x):
        """
        Calculate lambd, the lagrangian multiplier, for the Confidence Weighted Mean Reversion
        update method.

        :param new_m: (float) Dot product of relative returns and current weights.
        :param new_v: (float) The variance of current relative returns.
        :param new_w: (float) The weighted variance of relative returns.
        :param mean_x: (np.array) Weighted average of the variance.
        :return: (float) Lagrangian multiplier for the problem.
        """
        # Expression to speed up calculations.
        expn = new_v - np.dot(mean_x, new_w) + self.theta ** 2 * new_v / 2

        # Calculate constants for quadratic equation.
        quad_a = expn ** 2 - self.theta ** 4 * new_v ** 2 / 4
        quad_b = 2 * (self.epsilon - new_m) * expn
        quad_c = (self.epsilon - new_m) ** 2 - self.theta ** 2 * new_v

        # Calculate lagrangian.
        if quad_b ** 2 - 4 * quad_a * quad_c < 0:
            lambd = np.max([-quad_c / quad_b, 0])
        else:
            lambd = np.max([(-quad_b + np.sqrt(quad_b ** 2 - 4 * quad_a * quad_c)) / (2 * quad_a),
                            (-quad_b - np.sqrt(quad_b ** 2 - 4 * quad_a * quad_c)) / (2 * quad_a),
                            -quad_c / quad_b, 0])
        # Bound lambda for computational issues.
        lambd = np.minimum(lambd, 10000)
        return lambd
